# Generated from SparkyRPKParser.g4 by ANTLR 4.13.0
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,50,268,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,1,0,3,0,20,8,0,1,0,1,0,1,0,4,0,25,8,0,11,0,12,
        0,26,1,0,4,0,30,8,0,11,0,12,0,31,1,0,4,0,35,8,0,11,0,12,0,36,1,0,
        5,0,40,8,0,10,0,12,0,43,9,0,1,0,1,0,1,1,1,1,1,1,1,1,3,1,51,8,1,1,
        1,3,1,54,8,1,1,1,3,1,57,8,1,1,1,3,1,60,8,1,1,1,3,1,63,8,1,1,1,3,
        1,66,8,1,1,1,3,1,69,8,1,1,1,3,1,72,8,1,1,1,3,1,75,8,1,1,1,3,1,78,
        8,1,1,1,3,1,81,8,1,1,1,3,1,84,8,1,1,1,3,1,87,8,1,1,1,3,1,90,8,1,
        1,1,3,1,93,8,1,1,1,3,1,96,8,1,1,1,3,1,99,8,1,1,1,3,1,102,8,1,1,1,
        3,1,105,8,1,1,1,3,1,108,8,1,1,1,3,1,111,8,1,1,1,3,1,114,8,1,1,1,
        3,1,117,8,1,1,1,3,1,120,8,1,1,1,3,1,123,8,1,1,1,1,1,3,1,127,8,1,
        1,2,1,2,1,2,3,2,132,8,2,1,2,3,2,135,8,2,1,2,3,2,138,8,2,1,2,3,2,
        141,8,2,1,2,3,2,144,8,2,1,2,3,2,147,8,2,1,2,3,2,150,8,2,1,2,3,2,
        153,8,2,1,2,3,2,156,8,2,1,2,3,2,159,8,2,1,2,3,2,162,8,2,1,2,3,2,
        165,8,2,1,2,3,2,168,8,2,1,2,3,2,171,8,2,1,2,3,2,174,8,2,1,2,3,2,
        177,8,2,1,2,3,2,180,8,2,1,2,3,2,183,8,2,1,2,3,2,186,8,2,1,2,3,2,
        189,8,2,1,2,3,2,192,8,2,1,2,1,2,3,2,196,8,2,1,2,4,2,199,8,2,11,2,
        12,2,200,1,3,1,3,1,3,1,3,4,3,207,8,3,11,3,12,3,208,1,3,5,3,212,8,
        3,10,3,12,3,215,9,3,1,3,1,3,1,4,1,4,1,4,1,4,1,4,4,4,224,8,4,11,4,
        12,4,225,1,4,5,4,229,8,4,10,4,12,4,232,9,4,1,4,1,4,1,5,1,5,1,5,1,
        5,1,5,1,5,4,5,242,8,5,11,5,12,5,243,1,5,5,5,247,8,5,10,5,12,5,250,
        9,5,1,5,1,5,1,6,4,6,255,8,6,11,6,12,6,256,1,6,3,6,260,8,6,1,6,1,
        6,1,7,1,7,1,8,1,8,1,8,0,0,9,0,2,4,6,8,10,12,14,16,0,4,2,0,1,1,10,
        12,1,1,18,18,1,0,3,6,2,0,3,4,13,16,325,0,19,1,0,0,0,2,46,1,0,0,0,
        4,128,1,0,0,0,6,202,1,0,0,0,8,218,1,0,0,0,10,235,1,0,0,0,12,254,
        1,0,0,0,14,263,1,0,0,0,16,265,1,0,0,0,18,20,5,18,0,0,19,18,1,0,0,
        0,19,20,1,0,0,0,20,41,1,0,0,0,21,40,3,2,1,0,22,40,3,4,2,0,23,25,
        3,6,3,0,24,23,1,0,0,0,25,26,1,0,0,0,26,24,1,0,0,0,26,27,1,0,0,0,
        27,40,1,0,0,0,28,30,3,8,4,0,29,28,1,0,0,0,30,31,1,0,0,0,31,29,1,
        0,0,0,31,32,1,0,0,0,32,40,1,0,0,0,33,35,3,10,5,0,34,33,1,0,0,0,35,
        36,1,0,0,0,36,34,1,0,0,0,36,37,1,0,0,0,37,40,1,0,0,0,38,40,5,18,
        0,0,39,21,1,0,0,0,39,22,1,0,0,0,39,24,1,0,0,0,39,29,1,0,0,0,39,34,
        1,0,0,0,39,38,1,0,0,0,40,43,1,0,0,0,41,39,1,0,0,0,41,42,1,0,0,0,
        42,44,1,0,0,0,43,41,1,0,0,0,44,45,5,0,0,1,45,1,1,0,0,0,46,47,7,0,
        0,0,47,48,5,30,0,0,48,50,5,31,0,0,49,51,5,32,0,0,50,49,1,0,0,0,50,
        51,1,0,0,0,51,53,1,0,0,0,52,54,5,33,0,0,53,52,1,0,0,0,53,54,1,0,
        0,0,54,56,1,0,0,0,55,57,5,22,0,0,56,55,1,0,0,0,56,57,1,0,0,0,57,
        59,1,0,0,0,58,60,5,23,0,0,59,58,1,0,0,0,59,60,1,0,0,0,60,62,1,0,
        0,0,61,63,5,24,0,0,62,61,1,0,0,0,62,63,1,0,0,0,63,65,1,0,0,0,64,
        66,5,25,0,0,65,64,1,0,0,0,65,66,1,0,0,0,66,68,1,0,0,0,67,69,5,34,
        0,0,68,67,1,0,0,0,68,69,1,0,0,0,69,71,1,0,0,0,70,72,5,35,0,0,71,
        70,1,0,0,0,71,72,1,0,0,0,72,74,1,0,0,0,73,75,5,36,0,0,74,73,1,0,
        0,0,74,75,1,0,0,0,75,77,1,0,0,0,76,78,5,37,0,0,77,76,1,0,0,0,77,
        78,1,0,0,0,78,80,1,0,0,0,79,81,5,40,0,0,80,79,1,0,0,0,80,81,1,0,
        0,0,81,83,1,0,0,0,82,84,5,41,0,0,83,82,1,0,0,0,83,84,1,0,0,0,84,
        86,1,0,0,0,85,87,5,38,0,0,86,85,1,0,0,0,86,87,1,0,0,0,87,89,1,0,
        0,0,88,90,5,39,0,0,89,88,1,0,0,0,89,90,1,0,0,0,90,92,1,0,0,0,91,
        93,5,42,0,0,92,91,1,0,0,0,92,93,1,0,0,0,93,95,1,0,0,0,94,96,5,26,
        0,0,95,94,1,0,0,0,95,96,1,0,0,0,96,98,1,0,0,0,97,99,5,27,0,0,98,
        97,1,0,0,0,98,99,1,0,0,0,99,101,1,0,0,0,100,102,5,28,0,0,101,100,
        1,0,0,0,101,102,1,0,0,0,102,104,1,0,0,0,103,105,5,29,0,0,104,103,
        1,0,0,0,104,105,1,0,0,0,105,107,1,0,0,0,106,108,5,43,0,0,107,106,
        1,0,0,0,107,108,1,0,0,0,108,110,1,0,0,0,109,111,5,44,0,0,110,109,
        1,0,0,0,110,111,1,0,0,0,111,113,1,0,0,0,112,114,5,45,0,0,113,112,
        1,0,0,0,113,114,1,0,0,0,114,116,1,0,0,0,115,117,5,46,0,0,116,115,
        1,0,0,0,116,117,1,0,0,0,117,119,1,0,0,0,118,120,5,47,0,0,119,118,
        1,0,0,0,119,120,1,0,0,0,120,122,1,0,0,0,121,123,5,48,0,0,122,121,
        1,0,0,0,122,123,1,0,0,0,123,124,1,0,0,0,124,126,5,50,0,0,125,127,
        5,18,0,0,126,125,1,0,0,0,126,127,1,0,0,0,127,3,1,0,0,0,128,129,5,
        2,0,0,129,131,5,31,0,0,130,132,5,32,0,0,131,130,1,0,0,0,131,132,
        1,0,0,0,132,134,1,0,0,0,133,135,5,33,0,0,134,133,1,0,0,0,134,135,
        1,0,0,0,135,137,1,0,0,0,136,138,5,22,0,0,137,136,1,0,0,0,137,138,
        1,0,0,0,138,140,1,0,0,0,139,141,5,23,0,0,140,139,1,0,0,0,140,141,
        1,0,0,0,141,143,1,0,0,0,142,144,5,24,0,0,143,142,1,0,0,0,143,144,
        1,0,0,0,144,146,1,0,0,0,145,147,5,25,0,0,146,145,1,0,0,0,146,147,
        1,0,0,0,147,149,1,0,0,0,148,150,5,34,0,0,149,148,1,0,0,0,149,150,
        1,0,0,0,150,152,1,0,0,0,151,153,5,35,0,0,152,151,1,0,0,0,152,153,
        1,0,0,0,153,155,1,0,0,0,154,156,5,36,0,0,155,154,1,0,0,0,155,156,
        1,0,0,0,156,158,1,0,0,0,157,159,5,37,0,0,158,157,1,0,0,0,158,159,
        1,0,0,0,159,161,1,0,0,0,160,162,5,40,0,0,161,160,1,0,0,0,161,162,
        1,0,0,0,162,164,1,0,0,0,163,165,5,41,0,0,164,163,1,0,0,0,164,165,
        1,0,0,0,165,167,1,0,0,0,166,168,5,38,0,0,167,166,1,0,0,0,167,168,
        1,0,0,0,168,170,1,0,0,0,169,171,5,39,0,0,170,169,1,0,0,0,170,171,
        1,0,0,0,171,173,1,0,0,0,172,174,5,42,0,0,173,172,1,0,0,0,173,174,
        1,0,0,0,174,176,1,0,0,0,175,177,5,26,0,0,176,175,1,0,0,0,176,177,
        1,0,0,0,177,179,1,0,0,0,178,180,5,27,0,0,179,178,1,0,0,0,179,180,
        1,0,0,0,180,182,1,0,0,0,181,183,5,28,0,0,182,181,1,0,0,0,182,183,
        1,0,0,0,183,185,1,0,0,0,184,186,5,29,0,0,185,184,1,0,0,0,185,186,
        1,0,0,0,186,188,1,0,0,0,187,189,5,47,0,0,188,187,1,0,0,0,188,189,
        1,0,0,0,189,191,1,0,0,0,190,192,5,48,0,0,191,190,1,0,0,0,191,192,
        1,0,0,0,192,193,1,0,0,0,193,195,5,50,0,0,194,196,5,18,0,0,195,194,
        1,0,0,0,195,196,1,0,0,0,196,198,1,0,0,0,197,199,3,12,6,0,198,197,
        1,0,0,0,199,200,1,0,0,0,200,198,1,0,0,0,200,201,1,0,0,0,201,5,1,
        0,0,0,202,203,5,10,0,0,203,204,5,4,0,0,204,206,5,4,0,0,205,207,3,
        14,7,0,206,205,1,0,0,0,207,208,1,0,0,0,208,206,1,0,0,0,208,209,1,
        0,0,0,209,213,1,0,0,0,210,212,3,16,8,0,211,210,1,0,0,0,212,215,1,
        0,0,0,213,211,1,0,0,0,213,214,1,0,0,0,214,216,1,0,0,0,215,213,1,
        0,0,0,216,217,7,1,0,0,217,7,1,0,0,0,218,219,5,11,0,0,219,220,5,4,
        0,0,220,221,5,4,0,0,221,223,5,4,0,0,222,224,3,14,7,0,223,222,1,0,
        0,0,224,225,1,0,0,0,225,223,1,0,0,0,225,226,1,0,0,0,226,230,1,0,
        0,0,227,229,3,16,8,0,228,227,1,0,0,0,229,232,1,0,0,0,230,228,1,0,
        0,0,230,231,1,0,0,0,231,233,1,0,0,0,232,230,1,0,0,0,233,234,7,1,
        0,0,234,9,1,0,0,0,235,236,5,12,0,0,236,237,5,4,0,0,237,238,5,4,0,
        0,238,239,5,4,0,0,239,241,5,4,0,0,240,242,3,14,7,0,241,240,1,0,0,
        0,242,243,1,0,0,0,243,241,1,0,0,0,243,244,1,0,0,0,244,248,1,0,0,
        0,245,247,3,16,8,0,246,245,1,0,0,0,247,250,1,0,0,0,248,246,1,0,0,
        0,248,249,1,0,0,0,249,251,1,0,0,0,250,248,1,0,0,0,251,252,7,1,0,
        0,252,11,1,0,0,0,253,255,3,14,7,0,254,253,1,0,0,0,255,256,1,0,0,
        0,256,254,1,0,0,0,256,257,1,0,0,0,257,259,1,0,0,0,258,260,3,16,8,
        0,259,258,1,0,0,0,259,260,1,0,0,0,260,261,1,0,0,0,261,262,7,1,0,
        0,262,13,1,0,0,0,263,264,7,2,0,0,264,15,1,0,0,0,265,266,7,3,0,0,
        266,17,1,0,0,0,63,19,26,31,36,39,41,50,53,56,59,62,65,68,71,74,77,
        80,83,86,89,92,95,98,101,104,107,110,113,116,119,122,126,131,134,
        137,140,143,146,149,152,155,158,161,164,167,170,173,176,179,182,
        185,188,191,195,200,208,213,225,230,243,248,256,259
    ]

class SparkyRPKParser ( Parser ):

    grammarFileName = "SparkyRPKParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "'w2'", "'w3'", 
                     "'w4'", "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "'Volume'", "'Fit RMS %'" ]

    symbolicNames = [ "<INVALID>", "Assignment", "W1", "Integer", "Float", 
                      "Real", "Real_vol", "SHARP_COMMENT", "EXCLM_COMMENT", 
                      "SMCLN_COMMENT", "Assignment_2d_ex", "Assignment_3d_ex", 
                      "Assignment_4d_ex", "Note_2d_ex", "Note_3d_ex", "Note_4d_ex", 
                      "Simple_name", "SPACE", "RETURN", "ENCLOSE_COMMENT", 
                      "SECTION_COMMENT", "LINE_COMMENT", "W1_Hz_LA", "W2_Hz_LA", 
                      "W3_Hz_LA", "W4_Hz_LA", "Lw1_Hz_LA", "Lw2_Hz_LA", 
                      "Lw3_Hz_LA", "Lw4_Hz_LA", "W1_LA", "W2_LA", "W3_LA", 
                      "W4_LA", "Dev_w1_LA", "Dev_w2_LA", "Dev_w3_LA", "Dev_w4_LA", 
                      "Dummy_H_LA", "Height_LA", "Volume_LA", "Dummy_Rms_LA", 
                      "S_N_LA", "Atom1_LA", "Atom2_LA", "Atom3_LA", "Atom4_LA", 
                      "Distance_LA", "Note_LA", "SPACE_LA", "RETURN_LA" ]

    RULE_sparky_rpk = 0
    RULE_data_label = 1
    RULE_data_label_wo_assign = 2
    RULE_peak_2d = 3
    RULE_peak_3d = 4
    RULE_peak_4d = 5
    RULE_peak_wo_assign = 6
    RULE_number = 7
    RULE_note = 8

    ruleNames =  [ "sparky_rpk", "data_label", "data_label_wo_assign", "peak_2d", 
                   "peak_3d", "peak_4d", "peak_wo_assign", "number", "note" ]

    EOF = Token.EOF
    Assignment=1
    W1=2
    Integer=3
    Float=4
    Real=5
    Real_vol=6
    SHARP_COMMENT=7
    EXCLM_COMMENT=8
    SMCLN_COMMENT=9
    Assignment_2d_ex=10
    Assignment_3d_ex=11
    Assignment_4d_ex=12
    Note_2d_ex=13
    Note_3d_ex=14
    Note_4d_ex=15
    Simple_name=16
    SPACE=17
    RETURN=18
    ENCLOSE_COMMENT=19
    SECTION_COMMENT=20
    LINE_COMMENT=21
    W1_Hz_LA=22
    W2_Hz_LA=23
    W3_Hz_LA=24
    W4_Hz_LA=25
    Lw1_Hz_LA=26
    Lw2_Hz_LA=27
    Lw3_Hz_LA=28
    Lw4_Hz_LA=29
    W1_LA=30
    W2_LA=31
    W3_LA=32
    W4_LA=33
    Dev_w1_LA=34
    Dev_w2_LA=35
    Dev_w3_LA=36
    Dev_w4_LA=37
    Dummy_H_LA=38
    Height_LA=39
    Volume_LA=40
    Dummy_Rms_LA=41
    S_N_LA=42
    Atom1_LA=43
    Atom2_LA=44
    Atom3_LA=45
    Atom4_LA=46
    Distance_LA=47
    Note_LA=48
    SPACE_LA=49
    RETURN_LA=50

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.0")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class Sparky_rpkContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(SparkyRPKParser.EOF, 0)

        def RETURN(self, i:int=None):
            if i is None:
                return self.getTokens(SparkyRPKParser.RETURN)
            else:
                return self.getToken(SparkyRPKParser.RETURN, i)

        def data_label(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SparkyRPKParser.Data_labelContext)
            else:
                return self.getTypedRuleContext(SparkyRPKParser.Data_labelContext,i)


        def data_label_wo_assign(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SparkyRPKParser.Data_label_wo_assignContext)
            else:
                return self.getTypedRuleContext(SparkyRPKParser.Data_label_wo_assignContext,i)


        def peak_2d(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SparkyRPKParser.Peak_2dContext)
            else:
                return self.getTypedRuleContext(SparkyRPKParser.Peak_2dContext,i)


        def peak_3d(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SparkyRPKParser.Peak_3dContext)
            else:
                return self.getTypedRuleContext(SparkyRPKParser.Peak_3dContext,i)


        def peak_4d(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SparkyRPKParser.Peak_4dContext)
            else:
                return self.getTypedRuleContext(SparkyRPKParser.Peak_4dContext,i)


        def getRuleIndex(self):
            return SparkyRPKParser.RULE_sparky_rpk

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSparky_rpk" ):
                listener.enterSparky_rpk(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSparky_rpk" ):
                listener.exitSparky_rpk(self)




    def sparky_rpk(self):

        localctx = SparkyRPKParser.Sparky_rpkContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_sparky_rpk)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 19
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,0,self._ctx)
            if la_ == 1:
                self.state = 18
                self.match(SparkyRPKParser.RETURN)


            self.state = 41
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 269318) != 0):
                self.state = 39
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,4,self._ctx)
                if la_ == 1:
                    self.state = 21
                    self.data_label()
                    pass

                elif la_ == 2:
                    self.state = 22
                    self.data_label_wo_assign()
                    pass

                elif la_ == 3:
                    self.state = 24 
                    self._errHandler.sync(self)
                    _alt = 1
                    while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                        if _alt == 1:
                            self.state = 23
                            self.peak_2d()

                        else:
                            raise NoViableAltException(self)
                        self.state = 26 
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input,1,self._ctx)

                    pass

                elif la_ == 4:
                    self.state = 29 
                    self._errHandler.sync(self)
                    _alt = 1
                    while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                        if _alt == 1:
                            self.state = 28
                            self.peak_3d()

                        else:
                            raise NoViableAltException(self)
                        self.state = 31 
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input,2,self._ctx)

                    pass

                elif la_ == 5:
                    self.state = 34 
                    self._errHandler.sync(self)
                    _alt = 1
                    while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                        if _alt == 1:
                            self.state = 33
                            self.peak_4d()

                        else:
                            raise NoViableAltException(self)
                        self.state = 36 
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input,3,self._ctx)

                    pass

                elif la_ == 6:
                    self.state = 38
                    self.match(SparkyRPKParser.RETURN)
                    pass


                self.state = 43
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 44
            self.match(SparkyRPKParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Data_labelContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def W1_LA(self):
            return self.getToken(SparkyRPKParser.W1_LA, 0)

        def W2_LA(self):
            return self.getToken(SparkyRPKParser.W2_LA, 0)

        def RETURN_LA(self):
            return self.getToken(SparkyRPKParser.RETURN_LA, 0)

        def Assignment(self):
            return self.getToken(SparkyRPKParser.Assignment, 0)

        def Assignment_2d_ex(self):
            return self.getToken(SparkyRPKParser.Assignment_2d_ex, 0)

        def Assignment_3d_ex(self):
            return self.getToken(SparkyRPKParser.Assignment_3d_ex, 0)

        def Assignment_4d_ex(self):
            return self.getToken(SparkyRPKParser.Assignment_4d_ex, 0)

        def W3_LA(self):
            return self.getToken(SparkyRPKParser.W3_LA, 0)

        def W4_LA(self):
            return self.getToken(SparkyRPKParser.W4_LA, 0)

        def W1_Hz_LA(self):
            return self.getToken(SparkyRPKParser.W1_Hz_LA, 0)

        def W2_Hz_LA(self):
            return self.getToken(SparkyRPKParser.W2_Hz_LA, 0)

        def W3_Hz_LA(self):
            return self.getToken(SparkyRPKParser.W3_Hz_LA, 0)

        def W4_Hz_LA(self):
            return self.getToken(SparkyRPKParser.W4_Hz_LA, 0)

        def Dev_w1_LA(self):
            return self.getToken(SparkyRPKParser.Dev_w1_LA, 0)

        def Dev_w2_LA(self):
            return self.getToken(SparkyRPKParser.Dev_w2_LA, 0)

        def Dev_w3_LA(self):
            return self.getToken(SparkyRPKParser.Dev_w3_LA, 0)

        def Dev_w4_LA(self):
            return self.getToken(SparkyRPKParser.Dev_w4_LA, 0)

        def Volume_LA(self):
            return self.getToken(SparkyRPKParser.Volume_LA, 0)

        def Dummy_Rms_LA(self):
            return self.getToken(SparkyRPKParser.Dummy_Rms_LA, 0)

        def Dummy_H_LA(self):
            return self.getToken(SparkyRPKParser.Dummy_H_LA, 0)

        def Height_LA(self):
            return self.getToken(SparkyRPKParser.Height_LA, 0)

        def S_N_LA(self):
            return self.getToken(SparkyRPKParser.S_N_LA, 0)

        def Lw1_Hz_LA(self):
            return self.getToken(SparkyRPKParser.Lw1_Hz_LA, 0)

        def Lw2_Hz_LA(self):
            return self.getToken(SparkyRPKParser.Lw2_Hz_LA, 0)

        def Lw3_Hz_LA(self):
            return self.getToken(SparkyRPKParser.Lw3_Hz_LA, 0)

        def Lw4_Hz_LA(self):
            return self.getToken(SparkyRPKParser.Lw4_Hz_LA, 0)

        def Atom1_LA(self):
            return self.getToken(SparkyRPKParser.Atom1_LA, 0)

        def Atom2_LA(self):
            return self.getToken(SparkyRPKParser.Atom2_LA, 0)

        def Atom3_LA(self):
            return self.getToken(SparkyRPKParser.Atom3_LA, 0)

        def Atom4_LA(self):
            return self.getToken(SparkyRPKParser.Atom4_LA, 0)

        def Distance_LA(self):
            return self.getToken(SparkyRPKParser.Distance_LA, 0)

        def Note_LA(self):
            return self.getToken(SparkyRPKParser.Note_LA, 0)

        def RETURN(self):
            return self.getToken(SparkyRPKParser.RETURN, 0)

        def getRuleIndex(self):
            return SparkyRPKParser.RULE_data_label

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterData_label" ):
                listener.enterData_label(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitData_label" ):
                listener.exitData_label(self)




    def data_label(self):

        localctx = SparkyRPKParser.Data_labelContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_data_label)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 46
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 7170) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 47
            self.match(SparkyRPKParser.W1_LA)
            self.state = 48
            self.match(SparkyRPKParser.W2_LA)
            self.state = 50
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==32:
                self.state = 49
                self.match(SparkyRPKParser.W3_LA)


            self.state = 53
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==33:
                self.state = 52
                self.match(SparkyRPKParser.W4_LA)


            self.state = 56
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==22:
                self.state = 55
                self.match(SparkyRPKParser.W1_Hz_LA)


            self.state = 59
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==23:
                self.state = 58
                self.match(SparkyRPKParser.W2_Hz_LA)


            self.state = 62
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==24:
                self.state = 61
                self.match(SparkyRPKParser.W3_Hz_LA)


            self.state = 65
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==25:
                self.state = 64
                self.match(SparkyRPKParser.W4_Hz_LA)


            self.state = 68
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==34:
                self.state = 67
                self.match(SparkyRPKParser.Dev_w1_LA)


            self.state = 71
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==35:
                self.state = 70
                self.match(SparkyRPKParser.Dev_w2_LA)


            self.state = 74
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==36:
                self.state = 73
                self.match(SparkyRPKParser.Dev_w3_LA)


            self.state = 77
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==37:
                self.state = 76
                self.match(SparkyRPKParser.Dev_w4_LA)


            self.state = 80
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==40:
                self.state = 79
                self.match(SparkyRPKParser.Volume_LA)


            self.state = 83
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==41:
                self.state = 82
                self.match(SparkyRPKParser.Dummy_Rms_LA)


            self.state = 86
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==38:
                self.state = 85
                self.match(SparkyRPKParser.Dummy_H_LA)


            self.state = 89
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==39:
                self.state = 88
                self.match(SparkyRPKParser.Height_LA)


            self.state = 92
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==42:
                self.state = 91
                self.match(SparkyRPKParser.S_N_LA)


            self.state = 95
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==26:
                self.state = 94
                self.match(SparkyRPKParser.Lw1_Hz_LA)


            self.state = 98
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==27:
                self.state = 97
                self.match(SparkyRPKParser.Lw2_Hz_LA)


            self.state = 101
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==28:
                self.state = 100
                self.match(SparkyRPKParser.Lw3_Hz_LA)


            self.state = 104
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==29:
                self.state = 103
                self.match(SparkyRPKParser.Lw4_Hz_LA)


            self.state = 107
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==43:
                self.state = 106
                self.match(SparkyRPKParser.Atom1_LA)


            self.state = 110
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==44:
                self.state = 109
                self.match(SparkyRPKParser.Atom2_LA)


            self.state = 113
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==45:
                self.state = 112
                self.match(SparkyRPKParser.Atom3_LA)


            self.state = 116
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==46:
                self.state = 115
                self.match(SparkyRPKParser.Atom4_LA)


            self.state = 119
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==47:
                self.state = 118
                self.match(SparkyRPKParser.Distance_LA)


            self.state = 122
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==48:
                self.state = 121
                self.match(SparkyRPKParser.Note_LA)


            self.state = 124
            self.match(SparkyRPKParser.RETURN_LA)
            self.state = 126
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,31,self._ctx)
            if la_ == 1:
                self.state = 125
                self.match(SparkyRPKParser.RETURN)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Data_label_wo_assignContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def W1(self):
            return self.getToken(SparkyRPKParser.W1, 0)

        def W2_LA(self):
            return self.getToken(SparkyRPKParser.W2_LA, 0)

        def RETURN_LA(self):
            return self.getToken(SparkyRPKParser.RETURN_LA, 0)

        def W3_LA(self):
            return self.getToken(SparkyRPKParser.W3_LA, 0)

        def W4_LA(self):
            return self.getToken(SparkyRPKParser.W4_LA, 0)

        def W1_Hz_LA(self):
            return self.getToken(SparkyRPKParser.W1_Hz_LA, 0)

        def W2_Hz_LA(self):
            return self.getToken(SparkyRPKParser.W2_Hz_LA, 0)

        def W3_Hz_LA(self):
            return self.getToken(SparkyRPKParser.W3_Hz_LA, 0)

        def W4_Hz_LA(self):
            return self.getToken(SparkyRPKParser.W4_Hz_LA, 0)

        def Dev_w1_LA(self):
            return self.getToken(SparkyRPKParser.Dev_w1_LA, 0)

        def Dev_w2_LA(self):
            return self.getToken(SparkyRPKParser.Dev_w2_LA, 0)

        def Dev_w3_LA(self):
            return self.getToken(SparkyRPKParser.Dev_w3_LA, 0)

        def Dev_w4_LA(self):
            return self.getToken(SparkyRPKParser.Dev_w4_LA, 0)

        def Volume_LA(self):
            return self.getToken(SparkyRPKParser.Volume_LA, 0)

        def Dummy_Rms_LA(self):
            return self.getToken(SparkyRPKParser.Dummy_Rms_LA, 0)

        def Dummy_H_LA(self):
            return self.getToken(SparkyRPKParser.Dummy_H_LA, 0)

        def Height_LA(self):
            return self.getToken(SparkyRPKParser.Height_LA, 0)

        def S_N_LA(self):
            return self.getToken(SparkyRPKParser.S_N_LA, 0)

        def Lw1_Hz_LA(self):
            return self.getToken(SparkyRPKParser.Lw1_Hz_LA, 0)

        def Lw2_Hz_LA(self):
            return self.getToken(SparkyRPKParser.Lw2_Hz_LA, 0)

        def Lw3_Hz_LA(self):
            return self.getToken(SparkyRPKParser.Lw3_Hz_LA, 0)

        def Lw4_Hz_LA(self):
            return self.getToken(SparkyRPKParser.Lw4_Hz_LA, 0)

        def Distance_LA(self):
            return self.getToken(SparkyRPKParser.Distance_LA, 0)

        def Note_LA(self):
            return self.getToken(SparkyRPKParser.Note_LA, 0)

        def RETURN(self):
            return self.getToken(SparkyRPKParser.RETURN, 0)

        def peak_wo_assign(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SparkyRPKParser.Peak_wo_assignContext)
            else:
                return self.getTypedRuleContext(SparkyRPKParser.Peak_wo_assignContext,i)


        def getRuleIndex(self):
            return SparkyRPKParser.RULE_data_label_wo_assign

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterData_label_wo_assign" ):
                listener.enterData_label_wo_assign(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitData_label_wo_assign" ):
                listener.exitData_label_wo_assign(self)




    def data_label_wo_assign(self):

        localctx = SparkyRPKParser.Data_label_wo_assignContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_data_label_wo_assign)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 128
            self.match(SparkyRPKParser.W1)
            self.state = 129
            self.match(SparkyRPKParser.W2_LA)
            self.state = 131
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==32:
                self.state = 130
                self.match(SparkyRPKParser.W3_LA)


            self.state = 134
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==33:
                self.state = 133
                self.match(SparkyRPKParser.W4_LA)


            self.state = 137
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==22:
                self.state = 136
                self.match(SparkyRPKParser.W1_Hz_LA)


            self.state = 140
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==23:
                self.state = 139
                self.match(SparkyRPKParser.W2_Hz_LA)


            self.state = 143
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==24:
                self.state = 142
                self.match(SparkyRPKParser.W3_Hz_LA)


            self.state = 146
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==25:
                self.state = 145
                self.match(SparkyRPKParser.W4_Hz_LA)


            self.state = 149
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==34:
                self.state = 148
                self.match(SparkyRPKParser.Dev_w1_LA)


            self.state = 152
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==35:
                self.state = 151
                self.match(SparkyRPKParser.Dev_w2_LA)


            self.state = 155
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==36:
                self.state = 154
                self.match(SparkyRPKParser.Dev_w3_LA)


            self.state = 158
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==37:
                self.state = 157
                self.match(SparkyRPKParser.Dev_w4_LA)


            self.state = 161
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==40:
                self.state = 160
                self.match(SparkyRPKParser.Volume_LA)


            self.state = 164
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==41:
                self.state = 163
                self.match(SparkyRPKParser.Dummy_Rms_LA)


            self.state = 167
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==38:
                self.state = 166
                self.match(SparkyRPKParser.Dummy_H_LA)


            self.state = 170
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==39:
                self.state = 169
                self.match(SparkyRPKParser.Height_LA)


            self.state = 173
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==42:
                self.state = 172
                self.match(SparkyRPKParser.S_N_LA)


            self.state = 176
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==26:
                self.state = 175
                self.match(SparkyRPKParser.Lw1_Hz_LA)


            self.state = 179
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==27:
                self.state = 178
                self.match(SparkyRPKParser.Lw2_Hz_LA)


            self.state = 182
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==28:
                self.state = 181
                self.match(SparkyRPKParser.Lw3_Hz_LA)


            self.state = 185
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==29:
                self.state = 184
                self.match(SparkyRPKParser.Lw4_Hz_LA)


            self.state = 188
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==47:
                self.state = 187
                self.match(SparkyRPKParser.Distance_LA)


            self.state = 191
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==48:
                self.state = 190
                self.match(SparkyRPKParser.Note_LA)


            self.state = 193
            self.match(SparkyRPKParser.RETURN_LA)
            self.state = 195
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==18:
                self.state = 194
                self.match(SparkyRPKParser.RETURN)


            self.state = 198 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 197
                self.peak_wo_assign()
                self.state = 200 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not ((((_la) & ~0x3f) == 0 and ((1 << _la) & 120) != 0)):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_2dContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Assignment_2d_ex(self):
            return self.getToken(SparkyRPKParser.Assignment_2d_ex, 0)

        def Float(self, i:int=None):
            if i is None:
                return self.getTokens(SparkyRPKParser.Float)
            else:
                return self.getToken(SparkyRPKParser.Float, i)

        def RETURN(self):
            return self.getToken(SparkyRPKParser.RETURN, 0)

        def EOF(self):
            return self.getToken(SparkyRPKParser.EOF, 0)

        def number(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SparkyRPKParser.NumberContext)
            else:
                return self.getTypedRuleContext(SparkyRPKParser.NumberContext,i)


        def note(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SparkyRPKParser.NoteContext)
            else:
                return self.getTypedRuleContext(SparkyRPKParser.NoteContext,i)


        def getRuleIndex(self):
            return SparkyRPKParser.RULE_peak_2d

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_2d" ):
                listener.enterPeak_2d(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_2d" ):
                listener.exitPeak_2d(self)




    def peak_2d(self):

        localctx = SparkyRPKParser.Peak_2dContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_peak_2d)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 202
            self.match(SparkyRPKParser.Assignment_2d_ex)
            self.state = 203
            self.match(SparkyRPKParser.Float)
            self.state = 204
            self.match(SparkyRPKParser.Float)
            self.state = 206 
            self._errHandler.sync(self)
            _alt = 1
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 205
                    self.number()

                else:
                    raise NoViableAltException(self)
                self.state = 208 
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,55,self._ctx)

            self.state = 213
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 122904) != 0):
                self.state = 210
                self.note()
                self.state = 215
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 216
            _la = self._input.LA(1)
            if not(_la==-1 or _la==18):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_3dContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Assignment_3d_ex(self):
            return self.getToken(SparkyRPKParser.Assignment_3d_ex, 0)

        def Float(self, i:int=None):
            if i is None:
                return self.getTokens(SparkyRPKParser.Float)
            else:
                return self.getToken(SparkyRPKParser.Float, i)

        def RETURN(self):
            return self.getToken(SparkyRPKParser.RETURN, 0)

        def EOF(self):
            return self.getToken(SparkyRPKParser.EOF, 0)

        def number(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SparkyRPKParser.NumberContext)
            else:
                return self.getTypedRuleContext(SparkyRPKParser.NumberContext,i)


        def note(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SparkyRPKParser.NoteContext)
            else:
                return self.getTypedRuleContext(SparkyRPKParser.NoteContext,i)


        def getRuleIndex(self):
            return SparkyRPKParser.RULE_peak_3d

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_3d" ):
                listener.enterPeak_3d(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_3d" ):
                listener.exitPeak_3d(self)




    def peak_3d(self):

        localctx = SparkyRPKParser.Peak_3dContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_peak_3d)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 218
            self.match(SparkyRPKParser.Assignment_3d_ex)
            self.state = 219
            self.match(SparkyRPKParser.Float)
            self.state = 220
            self.match(SparkyRPKParser.Float)
            self.state = 221
            self.match(SparkyRPKParser.Float)
            self.state = 223 
            self._errHandler.sync(self)
            _alt = 1
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 222
                    self.number()

                else:
                    raise NoViableAltException(self)
                self.state = 225 
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,57,self._ctx)

            self.state = 230
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 122904) != 0):
                self.state = 227
                self.note()
                self.state = 232
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 233
            _la = self._input.LA(1)
            if not(_la==-1 or _la==18):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_4dContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Assignment_4d_ex(self):
            return self.getToken(SparkyRPKParser.Assignment_4d_ex, 0)

        def Float(self, i:int=None):
            if i is None:
                return self.getTokens(SparkyRPKParser.Float)
            else:
                return self.getToken(SparkyRPKParser.Float, i)

        def RETURN(self):
            return self.getToken(SparkyRPKParser.RETURN, 0)

        def EOF(self):
            return self.getToken(SparkyRPKParser.EOF, 0)

        def number(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SparkyRPKParser.NumberContext)
            else:
                return self.getTypedRuleContext(SparkyRPKParser.NumberContext,i)


        def note(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SparkyRPKParser.NoteContext)
            else:
                return self.getTypedRuleContext(SparkyRPKParser.NoteContext,i)


        def getRuleIndex(self):
            return SparkyRPKParser.RULE_peak_4d

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_4d" ):
                listener.enterPeak_4d(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_4d" ):
                listener.exitPeak_4d(self)




    def peak_4d(self):

        localctx = SparkyRPKParser.Peak_4dContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_peak_4d)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 235
            self.match(SparkyRPKParser.Assignment_4d_ex)
            self.state = 236
            self.match(SparkyRPKParser.Float)
            self.state = 237
            self.match(SparkyRPKParser.Float)
            self.state = 238
            self.match(SparkyRPKParser.Float)
            self.state = 239
            self.match(SparkyRPKParser.Float)
            self.state = 241 
            self._errHandler.sync(self)
            _alt = 1
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 240
                    self.number()

                else:
                    raise NoViableAltException(self)
                self.state = 243 
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,59,self._ctx)

            self.state = 248
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 122904) != 0):
                self.state = 245
                self.note()
                self.state = 250
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 251
            _la = self._input.LA(1)
            if not(_la==-1 or _la==18):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Peak_wo_assignContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def RETURN(self):
            return self.getToken(SparkyRPKParser.RETURN, 0)

        def EOF(self):
            return self.getToken(SparkyRPKParser.EOF, 0)

        def number(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SparkyRPKParser.NumberContext)
            else:
                return self.getTypedRuleContext(SparkyRPKParser.NumberContext,i)


        def note(self):
            return self.getTypedRuleContext(SparkyRPKParser.NoteContext,0)


        def getRuleIndex(self):
            return SparkyRPKParser.RULE_peak_wo_assign

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPeak_wo_assign" ):
                listener.enterPeak_wo_assign(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPeak_wo_assign" ):
                listener.exitPeak_wo_assign(self)




    def peak_wo_assign(self):

        localctx = SparkyRPKParser.Peak_wo_assignContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_peak_wo_assign)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 254 
            self._errHandler.sync(self)
            _alt = 1
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 253
                    self.number()

                else:
                    raise NoViableAltException(self)
                self.state = 256 
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,61,self._ctx)

            self.state = 259
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & 122904) != 0):
                self.state = 258
                self.note()


            self.state = 261
            _la = self._input.LA(1)
            if not(_la==-1 or _la==18):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class NumberContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Real_vol(self):
            return self.getToken(SparkyRPKParser.Real_vol, 0)

        def Real(self):
            return self.getToken(SparkyRPKParser.Real, 0)

        def Float(self):
            return self.getToken(SparkyRPKParser.Float, 0)

        def Integer(self):
            return self.getToken(SparkyRPKParser.Integer, 0)

        def getRuleIndex(self):
            return SparkyRPKParser.RULE_number

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNumber" ):
                listener.enterNumber(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNumber" ):
                listener.exitNumber(self)




    def number(self):

        localctx = SparkyRPKParser.NumberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_number)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 263
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 120) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class NoteContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Simple_name(self):
            return self.getToken(SparkyRPKParser.Simple_name, 0)

        def Integer(self):
            return self.getToken(SparkyRPKParser.Integer, 0)

        def Float(self):
            return self.getToken(SparkyRPKParser.Float, 0)

        def Note_2d_ex(self):
            return self.getToken(SparkyRPKParser.Note_2d_ex, 0)

        def Note_3d_ex(self):
            return self.getToken(SparkyRPKParser.Note_3d_ex, 0)

        def Note_4d_ex(self):
            return self.getToken(SparkyRPKParser.Note_4d_ex, 0)

        def getRuleIndex(self):
            return SparkyRPKParser.RULE_note

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNote" ):
                listener.enterNote(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNote" ):
                listener.exitNote(self)




    def note(self):

        localctx = SparkyRPKParser.NoteContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_note)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 265
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 122904) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








##
# File: CifReader.py derived from wwpdb.apps.ccmodule.io.ChemCompIo.py
# Date: 31-May-2010  John Westbrook
#
# Update:
# 06-Aug-2010 - jdw - generalized construction of methods to apply to any category
#                     add accessors for lists of dictionaries
# 12-May-2011 - rps - added check for None when asking for category Object in __getDataList()
# 24-Oct-2012 - rps - updated to reflect reorganization of modules in pdbx packages
# 23-Jul-2019 - my  - forked original code to wwpdb.util.nmr.CifReader
# 30-Jul-2019 - my  - add 'range-float' as filter item type
# 05-Aug-2019 - my  - add 'enum' as filter item type
# 28-Jan-2020 - my  - add 'withStructConf' option of getPolymerSequence
# 19-Mar-2020 - my  - add hasItem()
# 24-Mar-2020 - my  - add 'identical_chain_id' in results of getPolymerSequence()
# 15-Apr-2020 - my  - add 'total_models' option of getPolymerSequence (DAOTHER-4060)
# 19-Apr-2020 - my  - add random rotation test for detection of non-superimposed models (DAOTHER-4060)
# 08-May-2020 - my  - make sure parse() is run only once (DAOTHER-5654)
# 20-Nov-2020 - my  - additional support for insertion code in getPolymerSequence() (DAOTHER-6128)
# 29-Jun-2021 - my  - add 'auth_chain_id', 'identical_auth_chain_id' in results of getPolymerSequence() if possible (DAOTHER-7108)
# 14-Jan-2022 - my  - precise RMSD calculation with domain and medoid model identification (DAOTHER-4060, 7544)
# 02-Feb-2022 - my  - add 'abs-int', 'abs-float', 'range-int', 'range-abs-int', 'range-abs-float' as filter item types and 'not_equal_to' range filter (NMR restraint remediation)
# 30-Mar-2022 - my  - add support for _atom_site.label_alt_id (DAOTHER-4060, 7544, NMR restraint remediation)
# 06-Apr-2022 - my  - add support for auth_comp_id (DAOTHER-7690)
# 04-Aug-2022 - my  - detect sequence gaps in auth_seq_id, 'gap_in_auth_seq' (NMR restraint remediation)
# 10-Feb-2023 - my  - add 'fetch_first_match' filter to process large assembly avoiding forced timeout (NMR restraint remediation)
# 14-Apr-2023 - my  - enable to use cache datablock (NMR restraint remediation)
# 19-Apr-2023 - my  - support multiple datablock (NMR restraint validation)
# 24-Apr-2023 - my  - add 'default' attribute for key items (NMR restraint validation)
# 18-Dec-2023 - my  - add calculate_uninstanced_coord() (DAOTHER-8945)
# 24-Jan-2024 - my  - add 'default-from' attribute for key/data items (D_1300043061)
# 21-Feb-2024 - my  - add support for discontinuous model_id (NMR restraint remediation, 2n6j)
# 07-Mar-2024 - my  - extract pdbx_poly_seq_scheme.auth_mon_id as alt_cmop_id to prevent sequence mismatch due to 5-letter CCD ID (DAOTHER-9158 vs D_1300043061)
# 20-Aug-2024 - my  - support truncated loop sequence in the model (DAOTHER-9644)
# 10-Sep-2024 - my  - ignore identical polymer sequence extensions within polynucleotide multiplexes (DAOTHER-9674)
# 18-Sep-2024 - my  - add 'starts-with-alnum' item type (DAOTHER-9694)
##
""" A collection of classes for parsing CIF files, extracting polymer sequence, and RMSD calculation.
"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook, Masashi Yokochi"
__email__ = "jwest@rcsb.rutgers.edu, yokochi@protein.osaka-u.ac.jp"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "1.0.3"

import sys
import os
import math
import random
import itertools
import hashlib
import collections
import re
import copy
import inspect
import pickle

import numpy as np

from operator import itemgetter
from typing import IO, List, Tuple, Optional

from mmcif.api.PdbxContainers import DataContainer
from mmcif.io.PdbxReader import PdbxReader

from sklearn.cluster import DBSCAN
from rmsd.calculate_rmsd import (NAMES_ELEMENT, centroid, check_reflections, rmsd,  # noqa: F401 pylint: disable=no-name-in-module, import-error, unused-import
                                 kabsch_rmsd, quaternion_rmsd,
                                 reorder_hungarian, reorder_brute, reorder_distance,
                                 quaternion_rotate)

try:
    from wwpdb.utils.nmr.AlignUtil import emptyValue
except ImportError:
    from nmr.AlignUtil import emptyValue


# must be one of kabsch_rmsd, quaternion_rmsd, None
ROTATION_METHOD = quaternion_rmsd
# must be one of reorder_hungarian, reorder_brute, reorder_distance, None
REORDER_METHOD = reorder_hungarian
# scan through reflections in planes (e.g. Y transformed to -Y -> X, -Y, Z) and axis changes,
# (e.g. X and Z coords exchanged -> Z, Y, X). This will affect stereo-chemistry
USE_REFLECTIONS = False
# scan through reflections in planes (e.g. Y transformed to -Y -> X, -Y, Z) and axis changes,
# (e.g. X and Z coords exchanged -> Z, Y, X). Stereo-chemistry will be kept
USE_REFLECTIONS_KEEP_STEREO = False
REORDER = False

LEN_MAJOR_ASYM_ID = 26

SYMBOLS_ELEMENT = {k.upper(): v for k, v in NAMES_ELEMENT.items()}

CARTN_DATA_ITEMS = [{'name': 'Cartn_x', 'type': 'float', 'alt_name': 'x'},
                    {'name': 'Cartn_y', 'type': 'float', 'alt_name': 'y'},
                    {'name': 'Cartn_z', 'type': 'float', 'alt_name': 'z'}
                    ]


def M(axis: list, theta: float) -> list:
    """ Return the rotation matrix associated with counterclockwise rotation
        about the given axis by theta radians.
    """

    axis = np.asarray(axis)
    axis = axis / math.sqrt(np.dot(axis, axis))
    a = math.cos(theta / 2.0)
    b, c, d = -axis * math.sin(theta / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d

    return np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                     [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                     [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])


def to_np_array(a: dict) -> list:
    """ Return Numpy array of a given Cartesian coordinate
        in {'x': float, 'y': float, 'z': float} format.
    """

    return np.asarray([a['x'], a['y'], a['z']], dtype=float)


def get_coordinates(p: list) -> [list, list]:
    """ Convert list of atoms for RMSD calculation.
        @return: a vector set of the coordinates
    """

    V = []

    for a in p:
        V.append(to_np_array(a))

    atoms = [SYMBOLS_ELEMENT[a['element']] for a in p]

    V = np.asarray(V)
    atoms = np.asarray(atoms)

    assert V.shape[0] == atoms.size

    return atoms, V


def calculate_rmsd(p: list, q: list) -> float:
    """ Calculate RMSD of two coordinates.
        @return: RMSD value
    """

    p_atoms, p_coord = get_coordinates(p)
    q_atoms, q_coord = get_coordinates(q)

    p_size = len(p_atoms)
    q_size = len(q_atoms)

    assert p_size > 0 and q_size > 0
    assert p_size == q_size
    assert p_coord.shape[0] == q_coord.shape[0]
    assert np.count_nonzero(p_atoms != q_atoms) == 0 or REORDER

    p_cent = centroid(p_coord)
    q_cent = centroid(q_coord)

    p_coord -= p_cent
    q_coord -= q_cent

    result_rmsd = None

    if USE_REFLECTIONS or USE_REFLECTIONS_KEEP_STEREO:
        if 'rmsd_method' in inspect.getfullargspec(check_reflections).args:  # API changed - detect
            result_rmsd, _, _, q_review = check_reflections(p_atoms, q_atoms,  # pylint: disable=unexpected-keyword-arg
                                                            p_coord, q_coord,
                                                            reorder_method=REORDER_METHOD,
                                                            rmsd_method=ROTATION_METHOD,
                                                            keep_stereo=USE_REFLECTIONS_KEEP_STEREO)
        else:
            result_rmsd, _, _, q_review = check_reflections(p_atoms, q_atoms,  # pylint: disable=unexpected-keyword-arg
                                                            p_coord, q_coord,
                                                            reorder_method=REORDER_METHOD,
                                                            rotation_method=ROTATION_METHOD,
                                                            keep_stereo=USE_REFLECTIONS_KEEP_STEREO)

    elif REORDER:
        q_review = REORDER_METHOD(p_atoms, q_atoms, p_coord, q_coord)
        q_coord = q_coord[q_review]
        q_atoms = q_atoms[q_review]
        assert np.count_nonzero(p_atoms != q_atoms) == 0

    if result_rmsd is not None:
        pass

    elif ROTATION_METHOD is None:
        result_rmsd = rmsd(p_coord, q_coord)

    else:
        result_rmsd = ROTATION_METHOD(p_coord, q_coord)

    return result_rmsd


def calculate_uninstanced_coord(p_coord: list, q_coord: list, s_coord: list) -> [list, float]:
    """ Calculate RMSD of two reference coordinates (p_coord, q_coord)
        and complement missing coordinate (s_coord). (DAOTHER-8945)
        @return: complemented coordinates, RMSD value
    """

    assert p_coord.shape[0] == q_coord.shape[0]

    p_cent = centroid(p_coord)
    q_cent = centroid(q_coord)

    p_coord -= p_cent
    s_coord -= p_cent
    q_coord -= q_cent

    rot = quaternion_rotate(p_coord, q_coord)
    p_coord = np.dot(p_coord, rot)

    s_coord = np.dot(s_coord, rot)
    s_coord += q_cent

    return s_coord, quaternion_rmsd(p_coord, q_coord)


class CifReader:
    """ Accessor methods for parsing CIF files, extracting polymer sequence, and RMSD calculation.
    """

    def __init__(self, verbose: bool = True, log: IO = sys.stdout,
                 use_cache: bool = True,
                 sub_dir_name_for_cache: str = '.'):
        self.__class_name__ = self.__class__.__name__
        self.__version__ = __version__

        self.__verbose = verbose
        self.__lfh = log
        self.__debug = False

        # whether to use cache file
        self.__use_cache = use_cache

        # sub directory name for cache file
        self.__sub_dir_name_for_cache = sub_dir_name_for_cache

        # the current file path
        self.__filePath = None

        # the current working directory path
        self.__dirPath = None

        # the datablock list
        self.__dBlockList = None

        # the current list of datablock names
        self.__dBlockNameList = None

        # the primary datablock
        self.__dBlock = None

        # the category name list
        self.__categoryNameList = None

        # hash code of current cif file
        self.__hashCode = None

        # cache file path
        self.__cachePath = None

        # preset values
        self.trueValue = ('true', 't', 'yes', 'y', '1')

        # allowed item types
        self.itemTypes = ('str', 'bool',
                          'int', 'range-int', 'abs-int', 'range-abs-int',
                          'float', 'range-float', 'abs-float', 'range-abs-float',
                          'enum', 'starts-with-alnum')

        # random rotation test for detection of non-superimposed models (DAOTHER-4060)
        self.__random_rotaion_test = False
        self.__single_model_rotation_test = True

        if self.__random_rotaion_test:
            self.__lfh.write(f"+{self.__class_name__}.__init__() ++ Warning  - Enabled random rotation test\n")
            self.__lfh.write(f"+{self.__class_name__}.__init__() ++ Warning  - Single model rotation test: {self.__single_model_rotation_test}\n")

        # clustering parameters for recognition of well-defined regions
        self.__min_features_for_clustering = 4
        self.__max_features_for_clustering = 10
        self.__min_samples_for_clustering = 2
        self.__max_samples_for_clustering = 6

        # minimum monomers for domain recognition
        self.__min_monomers_for_domain = 12

        assert self.__min_monomers_for_domain > 6  # must be greater than 6 to prevent the 6xHIS tag from being recognized as a well-defined region

        # criterion for cutoff RMSD value
        self.__rmsd_cutoff = 3.5

        self.__d_cutoff = self.__rmsd_cutoff ** 2

        # criterion for detection of exactly overlaid models
        self.__rmsd_overlaid_exactly = 0.01

    def parse(self, filePath: str, dirPath: Optional[str] = None) -> bool:
        """ Parse CIF file, and set internal active datablock if possible.
            @return: True for success or False otherwise
        """

        if dirPath is not None:
            if os.path.isdir(dirPath):
                self.__dirPath = dirPath

        if self.__dBlock is not None and self.__filePath == filePath:
            return True

        self.__dBlockList = None
        self.__dBlockNameList = None
        self.__categoryNameList = None

        self.__dBlock = None
        self.__hashCode = None

        self.__filePath = filePath

        try:

            if not os.access(self.__filePath, os.R_OK):
                if self.__verbose:
                    self.__lfh.write(f"+{self.__class_name__} ++ Error  - Missing file {self.__filePath}\n")
                return False

            if self.__use_cache:
                with open(self.__filePath, 'r', encoding='utf-8', errors='ignore') as ifh:
                    self.__hashCode = hashlib.md5(ifh.read().encode('utf-8')).hexdigest()

            return self.__setDataBlock(self.__getDataBlockFromFile())

        except Exception as e:
            if self.__verbose and 'loop_ declaration outside of data_ block or save_ frame' not in str(e):
                self.__lfh.write(f"+{self.__class_name__} ++ Error  - Parse {self.__filePath} failed {str(e)}\n")
            return False

    def __getDataBlockFromFile(self, blockName: Optional[str] = None) -> Optional[DataContainer]:
        """ Worker method to read CIF file and set the target datablock.
            If no blockName is provided return the first datablock.
            @return: target datablock
        """

        if self.__use_cache:

            if self.__dirPath is None:
                self.__dirPath = os.path.dirname(self.__filePath)

            cache_dir = os.path.join(self.__dirPath, self.__sub_dir_name_for_cache)

            if not os.path.isdir(cache_dir):
                os.makedirs(cache_dir)

            self.__cachePath = os.path.join(cache_dir, f"{self.__hashCode}{'' if blockName is None else '_' + blockName}.pkl")

            if os.path.exists(self.__cachePath):

                try:

                    with open(self.__cachePath, 'rb') as ifh:
                        dBlock = pickle.load(ifh)
                        if dBlock is not None:
                            return dBlock

                    os.remove(self.__cachePath)

                except Exception:
                    pass

        if self.__dBlockList is None:
            self.__dBlockList, self.__dBlockNameList = [], []
            self.__categoryNameList = {}

            with open(self.__filePath, 'r', encoding='utf-8') as ifh:
                pRd = PdbxReader(ifh)
                pRd.read(self.__dBlockList)

                is_star = all(container.getType() == 'data' for container in self.__dBlockList)
                for container in self.__dBlockList:
                    if is_star or (not is_star and container.getType() != 'data'):
                        blockName = container.getName()
                        self.__dBlockNameList.append(blockName)

        if blockName is not None:
            for dBlock in self.__dBlockList:
                if dBlock.getType() == 'data' and dBlock.getName() == blockName:
                    return dBlock

        else:
            for dBlock in self.__dBlockList:
                if dBlock.getType() == 'data':
                    return dBlock

        return None

    def __setDataBlock(self, dataBlock: Optional[DataContainer] = None) -> bool:
        """ Assigns the input datablock as the active internal datablock.
            @return: True for success or False otherwise
        """

        try:

            if dataBlock.getType() == 'data':
                self.__dBlock = dataBlock

                if self.__use_cache and not os.path.exists(self.__cachePath):
                    with open(self.__cachePath, 'wb') as ofh:
                        pickle.dump(dataBlock, ofh)

                return True

        except Exception:
            pass

        self.__dBlock = None

        return False

    def getDirPath(self) -> str:
        """ Return directory path of CIF file.
        """

        return self.__dirPath

    def getFilePath(self) -> str:
        """ Return CIF file path.
        """

        return self.__filePath

    def getHashCode(self) -> str:
        """ Return hash code of the CIF file.
        """

        if self.__hashCode is None:
            with open(self.__filePath, 'r', encoding='utf-8', errors='ignore') as ifh:
                self.__hashCode = hashlib.md5(ifh.read().encode('utf-8')).hexdigest()

        return self.__hashCode

    def getDataBlockList(self) -> List[DataContainer]:
        """ Return list of datablocks.
        """

        return self.__dBlockList

    def getDataBlockNameList(self) -> List[str]:
        """ Return list of datablock names.
        """

        return self.__dBlockNameList

    def getDataBlock(self, blockName: Optional[str] = None) -> Optional[DataContainer]:
        """ Return target datablock.
            Return None in case current blockName does not exist or no blockName does not match.
            @return: target datablock
        """

        if self.__dBlock is None or self.__dBlock.getType() != 'data':
            return None

        if blockName is None or self.__dBlock.getName() == blockName:
            return self.__dBlock

        dBlock = self.__getDataBlockFromFile(blockName)

        return dBlock if self.__setDataBlock(dBlock) else None

    def getCategoryNameList(self, blockName: Optional[str] = None) -> List[str]:
        """ Return all category names in a given datablock.
        """

        if blockName is not None and self.__dBlock is not None and self.__dBlock.getName() != blockName:
            self.__setDataBlock(self.getDataBlock(blockName))

        if self.__dBlock is None:
            return []

        if self.__categoryNameList is None:
            self.__categoryNameList = {}

        if blockName in self.__categoryNameList:
            return self.__categoryNameList[blockName]

        try:

            categoryNameList = self.__dBlock.getObjNameList()

            return categoryNameList

        finally:
            self.__categoryNameList[blockName] = categoryNameList

    def hasCategory(self, catName: str, blockName: Optional[str] = None) -> bool:
        """ Return whether a given category exists.
        """

        return catName in self.getCategoryNameList(blockName)

    def hasItem(self, catName: str, itName: str, blockName: Optional[str] = None) -> bool:
        """ Return whether a given item exists in a category.
        """

        if blockName is not None and self.__dBlock is not None and self.__dBlock.getName() != blockName:
            self.__setDataBlock(self.getDataBlock(blockName))

        if self.__dBlock is None:
            return False

        catObj = self.__dBlock.getObj(catName)

        if catObj is None:
            return False

        return itName in catObj.getAttributeList()

    def getAttributeList(self, catName: str, blockName: Optional[str] = None) -> List[str]:
        """ Return item names of a given category.
        """

        if blockName is not None and self.__dBlock is not None and self.__dBlock.getName() != blockName:
            self.__setDataBlock(self.getDataBlock(blockName))

        if self.__dBlock is None:
            return []

        catObj = self.__dBlock.getObj(catName)

        if catObj is None:
            return []

        return catObj.getAttributeList()

    def getRowLength(self, catName: str, blockName: Optional[str] = None) -> int:
        """ Return length of rows of a given category.
        """

        if blockName is not None and self.__dBlock is not None and self.__dBlock.getName() != blockName:
            self.__setDataBlock(self.getDataBlock(blockName))

        if self.__dBlock is None:
            return 0

        catObj = self.__dBlock.getObj(catName)

        if catObj is not None:
            return len(catObj.getRowList())

        return 0

    def getRowList(self, catName: str, blockName: Optional[str] = None) -> List[list]:
        """ Return length of rows of a given category.
        """

        if blockName is not None and self.__dBlock is not None and self.__dBlock.getName() != blockName:
            self.__setDataBlock(self.getDataBlock(blockName))

        if self.__dBlock is None:
            return []

        catObj = self.__dBlock.getObj(catName)

        if catObj is not None:
            return catObj.getRowList()

        return []

    def getDictList(self, catName: str, blockName: Optional[str] = None) -> List[dict]:
        """ Return a list of dictionaries of a given category.
        """

        if blockName is not None and self.__dBlock is not None and self.__dBlock.getName() != blockName:
            self.__setDataBlock(self.getDataBlock(blockName))

        if self.__dBlock is None:
            return []

        catObj = self.__dBlock.getObj(catName)

        if catObj is None:
            return []

        iList = catObj.getAttributeList()

        dList = []
        for row in catObj.getRowList():
            dList.append({itName: row[idxIt] for idxIt, itName in enumerate(iList)})

        return dList

    def getDictListWithFilter(self, catName: str, dataItems: List[dict], filterItems: Optional[List[dict]] = None,
                              blockName: Optional[str] = None) -> List[dict]:
        """ Return a list of dictionaries of a given category with filter.
        """

        dataNames = [d['name'] for d in dataItems]

        for d in dataItems:
            if d['type'] not in self.itemTypes:
                raise TypeError(f"Type {d['type']} of data item {d['name']} must be one of {self.itemTypes}.")

        filterNames = None
        if filterItems is not None:
            filterNames = [f['name'] for f in filterItems]

            for f in filterItems:
                if f['type'] not in self.itemTypes:
                    raise TypeError(f"Type {f['type']} of filter item {f['name']} must be one of {self.itemTypes}.")

        if blockName is not None and self.__dBlock is not None and self.__dBlock.getName() != blockName:
            self.__setDataBlock(self.getDataBlock(blockName))

        if self.__dBlock is None:
            return []

        catObj = self.__dBlock.getObj(catName)

        if catObj is None:
            return []

        # get column name index
        colDict, fcolDict, fetchDict = {}, {}, {}  # 'fetch_first_match': True

        iList = catObj.getAttributeList()

        for idxIt, itName in enumerate(iList):
            if itName in dataNames:
                colDict[itName] = idxIt
            if filterNames is not None and itName in filterNames:
                fcolDict[itName] = idxIt

        if set(dataNames) & set(iList) != set(dataNames):
            raise LookupError(f"Missing one of data items {dataNames}.")

        if filterItems is not None and set(filterNames) & set(iList) != set(filterNames):
            raise LookupError(f"Missing one of filter items {filterNames}.")

        abort = False

        dList = []
        for row in catObj.getRowList():
            keep = True
            if filterItems is not None:
                for filterItem in filterItems:
                    name = filterItem['name']
                    val = row[fcolDict[name]]
                    if val in emptyValue:
                        if 'value' in filterItem and filterItem['value'] not in emptyValue:
                            keep = False
                            break
                    else:
                        filterItemType = filterItem['type']
                        if filterItemType in ('str', 'enum'):
                            pass
                        elif filterItemType == 'starts-with-alnum':
                            if not val[0].isalnum():
                                keep = False
                                break
                        elif filterItemType == 'bool':
                            val = val.lower() in self.trueValue
                        elif filterItemType == 'int':
                            try:
                                val = int(val)
                            except ValueError:
                                keep = False
                                break
                        elif filterItemType == 'float':
                            try:
                                val = float(val)
                            except ValueError:
                                keep = False
                                break
                        elif filterItemType in ('abs-int', 'range-abs-int'):
                            try:
                                val = abs(int(val))
                            except ValueError:
                                keep = False
                                break
                        else:  # 'range-float', 'range-abs-float'
                            try:
                                val = abs(float(val))
                            except ValueError:
                                keep = False
                                break
                        if filterItemType in ('range-int', 'range-abs-int', 'range-float', 'range-abs-float'):
                            _range = filterItem['range']
                            if ('min_exclusive' in _range and val <= _range['min_exclusive'])\
                               or ('min_inclusive' in _range and val < _range['min_inclusive'])\
                               or ('max_inclusive' in _range and val > _range['max_inclusive'])\
                               or ('max_exclusive' in _range and val >= _range['max_exclusive'])\
                               or ('not_equal_to' in _range and val == _range['not_equal_to']):
                                keep = False
                                break
                        elif filterItemType == 'enum':
                            if val not in filterItem['enum']:
                                keep = False
                                break
                            if 'fetch_first_match' in filterItem and filterItem['fetch_first_match']:
                                if name not in fetchDict:
                                    fetchDict[name] = val
                                elif val != fetchDict[name]:
                                    keep = False
                                    abort = True
                                    break
                        else:
                            if val != filterItem['value']:
                                keep = False
                                break
                            if 'fetch_first_match' in filterItem and filterItem['fetch_first_match']:
                                if name not in fetchDict:
                                    fetchDict[name] = val
                                elif val != fetchDict[name]:
                                    keep = False
                                    abort = True
                                    break

            if keep:
                tD = {}
                for dataItem in dataItems:
                    val = row[colDict[dataItem['name']]]
                    if val in emptyValue:
                        if 'default-from' in dataItem and dataItem['default-from'] in colDict:
                            val = row[colDict[dataItem['default-from']]]
                        else:
                            val = dataItem.get('default')
                    dataItemType = dataItem['type']
                    if dataItemType in ('str', 'enum'):
                        pass
                    elif dataItemType == 'starts-with-alnum':
                        if not val[0].isalnum():
                            val = None
                    elif dataItemType == 'bool':
                        val = val.lower() in self.trueValue
                    elif dataItemType == 'int' and val is not None:
                        try:
                            val = int(val)
                        except ValueError:
                            val = None
                    elif val is not None:
                        val = float(val)
                    if 'alt_name' in dataItem:
                        tD[dataItem['alt_name']] = val
                    else:
                        tD[dataItem['name']] = val
                dList.append(tD)

            elif abort:
                break

        return dList

    def getPolymerSequence(self, catName: str, keyItems: List[dict],
                           withStructConf: bool = False, withRmsd: bool = False, alias: bool = False,
                           totalModels: int = 1, effModelIds: Optional[List[int]] = None, repAltId: str = 'A') -> List[dict]:
        """ Extract sequence from a given loop in a CIF file.
        """

        keyNames = [k['name'] for k in keyItems]

        lenKeyItems = len(keyItems)

        if self.__dBlock is None:
            return []

        repModelId = effModelIds[0] if effModelIds is not None else 1

        # DAOTHER-9644: support for truncated loop in the model
        if withRmsd and catName == 'pdbx_poly_seq_scheme':  # avoid interference of ParserListenerUtils.coordAssemblyChecker()
            misPolyLink = []

            _catName = 'pdbx_validate_polymer_linkage'

            if self.hasCategory(_catName):
                _keyItems = [{'name': 'auth_asym_id_1', 'type': 'str', 'alt_name': 'auth_chain_id'},
                             {'name': 'auth_seq_id_1', 'type': 'int'},
                             {'name': 'auth_asym_id_2', 'type': 'str', 'alt_name': 'test_auth_chain_id'},
                             {'name': 'auth_seq_id_2', 'type': 'int'}
                             ]
                _filterItems = [{'name': 'PDB_model_num', 'type': 'int', 'value': repModelId},
                                {'name': 'label_alt_id_1', 'type': 'enum', 'enum': (repAltId,)},
                                {'name': 'label_alt_id_2', 'type': 'enum', 'enum': (repAltId,)}
                                ]

                for mis in self.getDictListWithFilter(_catName, _keyItems, _filterItems):
                    if mis['auth_chain_id'] == mis['test_auth_chain_id']:
                        del mis['test_auth_chain_id']
                        misPolyLink.append(mis)

        catObj = self.__dBlock.getObj(catName)

        if catObj is None:
            return []

        # get column name index
        itDict, altDict = {}, {}

        iList = catObj.getAttributeList()

        for idxIt, itName in enumerate(iList):
            itDict[itName] = idxIt
            if itName in keyNames:
                altDict[next(k['alt_name'] if 'alt_name' in k else itName for k in keyItems if k['name'] == itName)] = idxIt

        if set(keyNames) & set(itDict.keys()) != set(keyNames):
            raise LookupError(f"Missing one of data items {keyNames}.")

        # get row list
        rowList = catObj.getRowList()
        _rowList = None
        unmapSeqIds, unmapAuthSeqIds, mapAuthSeqIds = {}, {}, {}
        chainIdWoDefault = set()

        entityPoly = self.getDictList('entity_poly')

        # DAOTHER-9674
        for row in rowList:
            for j in range(lenKeyItems):
                itCol = itDict[keyNames[j]]
                if itCol < len(row) and row[itCol] in emptyValue:
                    if 'default-from' in keyItems[j] and keyItems[j]['default-from'] in keyNames:
                        if catName == 'pdbx_poly_seq_scheme':
                            if 'alt_name' in keyItems[j] and keyItems[j]['alt_name'] == 'auth_comp_id':
                                c = row[altDict['chain_id']]
                                etype = next((e['type'] for e in entityPoly if 'pdbx_strand_id' in e and c in e['pdbx_strand_id'].split(',')), None)
                                if etype is not None and 'polypeptide' not in etype:
                                    if c not in unmapSeqIds:
                                        unmapSeqIds[c], unmapAuthSeqIds[c] = [], []
                                    compId = row[altDict['comp_id']]
                                    if compId in emptyValue or not compId[0].isalnum():  # DAOTHER-9694
                                        continue
                                    unmapSeqIds[c].append((row[altDict['seq_id']], compId))
                                    unmapAuthSeqIds[c].append(row[altDict['auth_seq_id']])
                                if _rowList is None:
                                    _rowList = copy.deepcopy(rowList)
                        continue
                    if 'default' not in keyItems[j] or keyItems[j]['default'] not in emptyValue:
                        raise ValueError(f"{keyNames[j]} must not be empty.")

        # DAOTHER-9674
        if catName == 'pdbx_poly_seq_scheme' and 'auth_comp_id' in altDict:
            for row in rowList:
                if row[altDict['auth_comp_id']] not in emptyValue:
                    c = row[altDict['chain_id']]
                    etype = next((e['type'] for e in entityPoly if 'pdbx_strand_id' in e and c in e['pdbx_strand_id'].split(',')), None)
                    if etype is not None and 'polypeptide' not in etype:
                        if c not in mapAuthSeqIds:
                            mapAuthSeqIds[c] = []
                        compId = row[altDict['comp_id']]
                        if compId in emptyValue or not compId[0].isalnum():  # DAOTHER-9694
                            continue
                        mapAuthSeqIds[c].append(row[altDict['auth_seq_id']])

        # DAOTHER-9674
        if len(unmapSeqIds) > 1:
            for (i, j) in itertools.combinations(unmapSeqIds.keys(), 2):
                if (i not in chainIdWoDefault or j not in chainIdWoDefault)\
                   and unmapSeqIds[i] == unmapSeqIds[j]\
                   and (len(unmapAuthSeqIds[i]) % len(mapAuthSeqIds[i]) == 0
                        or len(mapAuthSeqIds[i]) % len(unmapAuthSeqIds[i]) == 0):
                    chainIdWoDefault.add(i)
                    chainIdWoDefault.add(j)

            if len(chainIdWoDefault) > 1:
                rowList = []

                for row in _rowList:
                    skip = False
                    for j in range(lenKeyItems):
                        itCol = itDict[keyNames[j]]
                        if itCol < len(row) and row[itCol] in emptyValue:
                            if 'default-from' in keyItems[j] and keyItems[j]['default-from'] in keyNames:
                                if catName == 'pdbx_poly_seq_scheme':
                                    if 'alt_name' in keyItems[j] and keyItems[j]['alt_name'] == 'auth_comp_id':
                                        c = row[altDict['chain_id']]
                                        if c in chainIdWoDefault:
                                            skip = True
                                            break
                                row[itCol] = row[itDict[keyItems[j]['default-from']]]
                    if not skip:
                        rowList.append(row)

        for row in rowList:
            for j in range(lenKeyItems):
                itCol = itDict[keyNames[j]]
                if itCol < len(row) and row[itCol] in emptyValue:
                    if 'default-from' in keyItems[j] and keyItems[j]['default-from'] in keyNames:
                        if catName == 'pdbx_poly_seq_scheme':
                            if 'alt_name' in keyItems[j] and keyItems[j]['alt_name'] == 'auth_comp_id':
                                c = row[altDict['chain_id']]
                        row[itCol] = row[itDict[keyItems[j]['default-from']]]
                        continue

        compDict, seqDict, insCodeDict, authSeqDict, labelSeqDict, authChainDict =\
            {}, {}, {}, {}, {}, {}

        chain_id_col = altDict['chain_id']
        seq_id_col = altDict['seq_id']
        comp_id_col = altDict['comp_id']
        ins_code_col = -1 if 'ins_code' not in altDict else altDict['ins_code']
        label_seq_col = seq_id_col if 'label_seq_id' not in altDict else altDict['label_seq_id']
        auth_chain_id_col = -1 if 'auth_chain_id' not in altDict else altDict['auth_chain_id']
        auth_seq_id_col = -1 if 'auth_seq_id' not in altDict else altDict['auth_seq_id']
        auth_comp_id_col = -1 if 'auth_comp_id' not in altDict else altDict['auth_comp_id']
        alt_comp_id_col = -1 if 'alt_comp_id' not in altDict else altDict['alt_comp_id']

        authScheme = auth_seq_id_col != -1

        chainIds = sorted(set(row[chain_id_col] for row in rowList), key=lambda x: (len(x), x))

        if ins_code_col == -1:
            if catName == 'pdbx_nonpoly_scheme':
                sortedSeq = sorted(set((row[chain_id_col], int(row[seq_id_col]), row[comp_id_col]) for row in rowList),
                                   key=itemgetter(1))
            else:
                sortedSeq = sorted(set((row[chain_id_col], int(row[seq_id_col]), row[comp_id_col]) for row in rowList),
                                   key=lambda x: (len(x[0]), x[0], x[1]))

            keyDict = {(row[chain_id_col], int(row[seq_id_col])): row[comp_id_col] for row in rowList}

            for row in rowList:
                key = (row[chain_id_col], int(row[seq_id_col]))
                if keyDict[key] != row[comp_id_col]:
                    raise KeyError(f"Sequence must be unique. {iList[chain_id_col]} {row[chain_id_col]}, "
                                   f"{iList[seq_id_col]} {row[seq_id_col]}, "
                                   f"{iList[comp_id_col]} {row[comp_id_col]} vs {keyDict[key]}.")

            for c in chainIds:
                compDict[c] = [x[2] for x in sortedSeq if x[0] == c]
                seqDict[c] = [x[1] for x in sortedSeq if x[0] == c]

        else:
            if catName == 'pdbx_nonpoly_scheme':
                sortedSeq = sorted(set((row[chain_id_col], int(row[seq_id_col]), row[ins_code_col], row[label_seq_col], row[comp_id_col]) for row in rowList),
                                   key=itemgetter(1))
            else:
                if all(row[label_seq_col].isdigit() for row in rowList):
                    sortedSeq = sorted(set((row[chain_id_col], int(row[seq_id_col]), row[ins_code_col], int(row[label_seq_col]), row[comp_id_col]) for row in rowList),
                                       key=lambda x: (len(x[0]), x[0], x[3]))
                else:
                    sortedSeq = sorted(set((row[chain_id_col], int(row[seq_id_col]), row[ins_code_col], row[label_seq_col], row[comp_id_col]) for row in rowList),
                                       key=lambda x: (len(x[0]), x[0], x[1]))

            keyDict = {(row[chain_id_col], int(row[seq_id_col]), row[ins_code_col], row[label_seq_col]): row[comp_id_col] for row in rowList}

            for row in rowList:
                key = (row[chain_id_col], int(row[seq_id_col]), row[ins_code_col], row[label_seq_col])
                if keyDict[key] != row[comp_id_col]:
                    raise KeyError(f"Sequence must be unique. {iList[chain_id_col]} {row[chain_id_col]}, "
                                   f"{iList[seq_id_col]} {row[seq_id_col]}, "
                                   f"{iList[ins_code_col]} {row[ins_code_col]}, "
                                   f"{iList[label_seq_col]} {row[label_seq_col]}, "
                                   f"{iList[comp_id_col]} {row[comp_id_col]} vs {keyDict[key]}.")

            for c in chainIds:
                compDict[c] = [x[4] for x in sortedSeq if x[0] == c]
                seqDict[c] = [x[1] for x in sortedSeq if x[0] == c]
                insCodeDict[c] = [x[2] for x in sortedSeq if x[0] == c]
                labelSeqDict[c] = [x[3] for x in sortedSeq if x[0] == c]

        chainIds = []
        for x in sortedSeq:
            if x[0] not in chainIds:
                chainIds.append(x[0])

        if auth_chain_id_col != -1:
            for row in rowList:
                c = row[chain_id_col]
                if c not in authChainDict:
                    authChainDict[c] = row[auth_chain_id_col]

        if authScheme:
            for c in chainIds:
                authSeqDict[c] = []
                for s in seqDict[c]:
                    row = next((row for row in rowList if row[chain_id_col] == c and int(row[seq_id_col]) == s), None)
                    if row is not None:
                        if row[auth_seq_id_col] not in emptyValue:
                            try:
                                _s = int(row[auth_seq_id_col])
                            except ValueError:
                                _s = None
                            authSeqDict[c].append(_s)
                        else:
                            authSeqDict[c].append(None)

        largeAssembly = catName == 'pdbx_poly_seq_scheme' and len(chainIds) > LEN_MAJOR_ASYM_ID

        caRmsd = caWellDefinedRegion = None
        polyPeptideChains, polyPeptideLengths = [], []

        _seqDict = copy.deepcopy(seqDict)

        asm = []  # assembly of a loop
        for i, c in enumerate(chainIds):
            ent = {}  # entity

            ident = False
            if len(asm) > 0 and largeAssembly:
                _ent = asm[-1]
                if 'identical_chain_id' in _ent and c in _ent['identical_chain_id']:
                    ident = True

            if ident:
                ent = copy.copy(asm[-1])

            ent['chain_id'] = ent['auth_chain_id'] = c
            if auth_chain_id_col != -1:
                ent['auth_chain_id'] = authChainDict[c]

            if not ident:

                etype = next((e['type'] for e in entityPoly if 'pdbx_strand_id' in e and c in e['pdbx_strand_id'].split(',')), None)

                # DAOTHER-9644: support for truncated loop in the model
                if withRmsd and catName == 'pdbx_poly_seq_scheme' and len(authSeqDict) > 0:  # avoid interference of ParserListenerUtils.coordAssemblyChecker()

                    if len(misPolyLink) > 0:

                        for mis in misPolyLink:

                            if mis['auth_chain_id'] != (authChainDict[c] if auth_chain_id_col != -1 else c):
                                continue

                            auth_seq_id_1 = mis['auth_seq_id_1']
                            auth_seq_id_2 = mis['auth_seq_id_2']

                            if auth_seq_id_1 in authSeqDict[c]\
                               and auth_seq_id_2 in authSeqDict[c]\
                               and auth_seq_id_1 < auth_seq_id_2:

                                for auth_seq_id_ in range(auth_seq_id_1 + 1, auth_seq_id_2):
                                    auth_seq_id_list = list(filter(None, authSeqDict[c]))

                                    if auth_seq_id_ < min(auth_seq_id_list):
                                        pos = 0
                                    elif auth_seq_id_ > max(auth_seq_id_list):
                                        pos = len(auth_seq_id_list)
                                    else:
                                        for idx, _auth_seq_id_ in enumerate(auth_seq_id_list):
                                            if _auth_seq_id_ < auth_seq_id_:
                                                continue
                                            pos = idx
                                            break

                                    authSeqDict[c].insert(pos, auth_seq_id_)
                                    compDict[c].insert(pos, '.')  # DAOTHER-9644: comp_id must be specified at Macromelucule page
                                    if ins_code_col != -1:
                                        insCodeDict[c].insert(pos, '.')

                                # DAOTHER-9644: insert label_seq_id for truncated loop in the coordinates
                                seqDict[c] = labelSeqDict[c] = list(range(1, len(authSeqDict[c]) + 1))

                    # DAOTHER-9644: simulate pdbx_poly_seq_scheme category
                    elif etype is not None:

                        if 'polypeptide' in etype:
                            BEG_ATOM = "C"
                            END_ATOM = "N"
                        else:
                            BEG_ATOM = "O3'"
                            END_ATOM = "P"

                        has_ins_code = False

                        for p in range(len(authSeqDict[c]) - 1):
                            s_p = authSeqDict[c][p]
                            s_q = authSeqDict[c][p + 1]

                            if None in (s_p, s_q):
                                continue

                            if s_p == s_q:
                                has_ins_code = True
                                continue

                            if s_p + 1 != s_q:

                                if has_ins_code:
                                    has_ins_code = False
                                    continue

                                auth_seq_id_1 = s_p
                                auth_seq_id_2 = s_q

                                _beg =\
                                    self.getDictListWithFilter('atom_site',
                                                               CARTN_DATA_ITEMS,
                                                               [{'name': 'label_asym_id', 'type': 'str', 'value': c},
                                                                {'name': 'auth_seq_id', 'type': 'int', 'value': auth_seq_id_1},
                                                                {'name': 'label_atom_id', 'type': 'starts-with-alnum', 'value': BEG_ATOM},
                                                                {'name': 'pdbx_PDB_model_num', 'type': 'int', 'value': repModelId},
                                                                {'name': 'label_alt_id', 'type': 'enum', 'enum': (repAltId,)}
                                                                ])

                                _end =\
                                    self.getDictListWithFilter('atom_site',
                                                               CARTN_DATA_ITEMS,
                                                               [{'name': 'label_asym_id', 'type': 'str', 'value': c},
                                                                {'name': 'auth_seq_id', 'type': 'int', 'value': auth_seq_id_2},
                                                                {'name': 'label_atom_id', 'type': 'starts-with-alnum', 'value': END_ATOM},
                                                                {'name': 'pdbx_PDB_model_num', 'type': 'int', 'value': repModelId},
                                                                {'name': 'label_alt_id', 'type': 'enum', 'enum': (repAltId,)}
                                                                ])

                                if len(_beg) == 1 and len(_end) == 1 and np.linalg.norm(to_np_array(_beg[0]) - to_np_array(_end[0])) > 5.0:
                                    for auth_seq_id_ in range(auth_seq_id_1 + 1, auth_seq_id_2):
                                        auth_seq_id_list = list(filter(None, authSeqDict[c]))

                                        if auth_seq_id_ < min(auth_seq_id_list):
                                            pos = 0
                                        elif auth_seq_id_ > max(auth_seq_id_list):
                                            pos = len(auth_seq_id_list)
                                        else:
                                            for idx, _auth_seq_id_ in enumerate(auth_seq_id_list):
                                                if _auth_seq_id_ < auth_seq_id_:
                                                    continue
                                                pos = idx
                                                break

                                        authSeqDict[c].insert(pos, auth_seq_id_)
                                        compDict[c].insert(pos, '.')  # DAOTHER-9644: comp_id must be specified at Macromelucule page
                                        if ins_code_col != -1:
                                            insCodeDict[c].insert(pos, '.')

                                    # DAOTHER-9644: insert label_seq_id for truncated loop in the coordinates
                                    seqDict[c] = labelSeqDict[c] = list(range(1, len(authSeqDict[c]) + 1))

                ent['seq_id'] = seqDict[c]
                ent['comp_id'] = compDict[c]
                if c in insCodeDict:
                    if any(ic for ic in insCodeDict[c] if ic not in emptyValue):
                        ent['ins_code'] = insCodeDict[c]
                    if any(s for s in labelSeqDict[c] if s not in emptyValue):
                        if c in labelSeqDict and all(isinstance(s, int) for s in labelSeqDict[c]):
                            ent['auth_seq_id'] = authSeqDict[c] if authScheme else seqDict[c]
                            ent['label_seq_id'] = labelSeqDict[c]
                            ent['seq_id'] = ent['label_seq_id']

                if authScheme:
                    ent['auth_seq_id'] = authSeqDict[c]
                    ent['gap_in_auth_seq'] = False
                    for p in range(len(authSeqDict[c]) - 1):
                        s_p = ent['auth_seq_id'][p]
                        s_q = ent['auth_seq_id'][p + 1]
                        if None in (s_p, s_q):
                            continue
                        if s_p + 1 != s_q:
                            ent['gap_in_auth_seq'] = True
                            break

                if auth_comp_id_col != -1:
                    ent['auth_comp_id'] = []
                    if authScheme:
                        if ins_code_col != -1:
                            for s, ic in zip(authSeqDict[c], insCodeDict[c]):
                                row = next((row for row in rowList if row[chain_id_col] == c
                                            and int(row[auth_seq_id_col]) == s and row[ins_code_col] == ic), None)
                                if row is not None:
                                    comp_id = row[auth_comp_id_col]
                                    if comp_id not in emptyValue:
                                        ent['auth_comp_id'].append(comp_id)
                                    else:
                                        ent['auth_comp_id'].append('.')
                                else:
                                    ent['auth_comp_id'].append('.')
                        else:
                            for s in authSeqDict[c]:
                                row = next((row for row in rowList if row[chain_id_col] == c
                                            and int(row[auth_seq_id_col]) == s), None)
                                if row is not None:
                                    comp_id = row[auth_comp_id_col]
                                    if comp_id not in emptyValue:
                                        ent['auth_comp_id'].append(comp_id)
                                    else:
                                        ent['auth_comp_id'].append('.')
                                else:
                                    ent['auth_comp_id'].append('.')
                    else:
                        for s in seqDict[c]:
                            row = next((row for row in rowList if row[chain_id_col] == c
                                        and int(row[seq_id_col]) == s), None)
                            if row is not None:
                                comp_id = row[auth_comp_id_col]
                                if comp_id not in emptyValue:
                                    ent['auth_comp_id'].append(comp_id)
                                else:
                                    ent['auth_comp_id'].append('.')
                            else:
                                ent['auth_comp_id'].append('.')
                else:
                    ent['auth_comp_id'] = ent['comp_id']

                if alt_comp_id_col != -1:
                    ent['alt_comp_id'] = []
                    if authScheme:
                        if ins_code_col != -1:
                            for s, ic in zip(authSeqDict[c], insCodeDict[c]):
                                row = next((row for row in rowList if row[chain_id_col] == c
                                            and int(row[auth_seq_id_col]) == s and row[ins_code_col] == ic), None)
                                if row is not None:
                                    comp_id = row[alt_comp_id_col]
                                    if comp_id not in emptyValue:
                                        ent['alt_comp_id'].append(comp_id)
                                    else:
                                        ent['alt_comp_id'].append('.')
                                else:
                                    ent['alt_comp_id'].append('.')
                        else:
                            for s in authSeqDict[c]:
                                row = next((row for row in rowList if row[chain_id_col] == c
                                            and int(row[auth_seq_id_col]) == s), None)
                                if row is not None:
                                    comp_id = row[alt_comp_id_col]
                                    if comp_id not in emptyValue:
                                        ent['alt_comp_id'].append(comp_id)
                                    else:
                                        ent['alt_comp_id'].append('.')
                                else:
                                    ent['alt_comp_id'].append('.')
                    else:
                        for s in seqDict[c]:
                            row = next((row for row in rowList if row[chain_id_col] == c
                                        and int(row[seq_id_col]) == s), None)
                            if row is not None:
                                comp_id = row[alt_comp_id_col]
                                if comp_id not in emptyValue:
                                    ent['alt_comp_id'].append(comp_id)
                                else:
                                    ent['alt_comp_id'].append('.')
                            else:
                                ent['alt_comp_id'].append('.')

                if withStructConf and i < LEN_MAJOR_ASYM_ID:  # to process large assembly avoiding forced timeout
                    ent['struct_conf'] = self.__extractStructConf(c, authSeqDict[c] if authScheme else seqDict[c], not authScheme)

                # to process large assembly avoiding forced timeout (2ms7, 21 chains)
                if withRmsd and etype is not None and totalModels > 1 and i < LEN_MAJOR_ASYM_ID / 2:
                    ent['type'] = etype

                    randomM = None
                    if self.__random_rotaion_test:
                        randomM = {}
                        for model_id in effModelIds:
                            axis = [random.uniform(-1.0, 1.0), random.uniform(-1.0, 1.0), random.uniform(-1.0, 1.0)]
                            if self.__single_model_rotation_test:
                                theta = 0.0 if model_id > 1 else np.pi / 4.0
                            else:
                                theta = random.uniform(-np.pi, np.pi)
                            randomM[model_id] = M(axis, theta)

                    if 'polypeptide' in etype:

                        if caRmsd is None:

                            polyPeptideChains = [c]
                            polyPeptideLengths = [len(_seqDict[c])]

                            for c2 in chainIds:

                                if c2 == c:
                                    continue

                                etype2 = next((e['type'] for e in entityPoly if 'pdbx_strand_id' in e and c2 in e['pdbx_strand_id'].split(',')), None)

                                if etype2 is not None and 'polypeptide' in etype2:
                                    polyPeptideChains.append(c2)
                                    polyPeptideLengths.append(len(_seqDict[c2]))

                            ca_atom_sites = self.getDictListWithFilter('atom_site',
                                                                       [{'name': 'Cartn_x', 'type': 'float', 'alt_name': 'x'},
                                                                        {'name': 'Cartn_y', 'type': 'float', 'alt_name': 'y'},
                                                                        {'name': 'Cartn_z', 'type': 'float', 'alt_name': 'z'},
                                                                        {'name': 'label_asym_id', 'type': 'str', 'alt_name': 'chain_id'},
                                                                        {'name': 'auth_seq_id', 'type': 'int', 'alt_name': 'seq_id'},
                                                                        {'name': 'ndb_model' if alias else 'pdbx_PDB_model_num', 'type': 'int', 'alt_name': 'model_id'},
                                                                        {'name': 'type_symbol', 'type': 'str', 'alt_name': 'element'}
                                                                        ],
                                                                       [{'name': 'label_asym_id', 'type': 'enum',
                                                                         'enum': polyPeptideChains},
                                                                        {'name': 'label_atom_id', 'type': 'starts-with-alnum', 'value': 'CA'},
                                                                        {'name': 'label_alt_id', 'type': 'enum',
                                                                         'enum': (repAltId,)},
                                                                        {'name': 'type_symbol', 'type': 'str', 'value': 'C'}])

                            co_atom_sites = self.getDictListWithFilter('atom_site',
                                                                       [{'name': 'Cartn_x', 'type': 'float', 'alt_name': 'x'},
                                                                        {'name': 'Cartn_y', 'type': 'float', 'alt_name': 'y'},
                                                                        {'name': 'Cartn_z', 'type': 'float', 'alt_name': 'z'},
                                                                        {'name': 'label_asym_id', 'type': 'str', 'alt_name': 'chain_id'},
                                                                        {'name': 'auth_seq_id', 'type': 'int', 'alt_name': 'seq_id'},
                                                                        {'name': 'ndb_model' if alias else 'pdbx_PDB_model_num', 'type': 'int', 'alt_name': 'model_id'},
                                                                        {'name': 'type_symbol', 'type': 'str', 'alt_name': 'element'}
                                                                        ],
                                                                       [{'name': 'label_asym_id', 'type': 'enum',
                                                                         'enum': polyPeptideChains},
                                                                        {'name': 'label_atom_id', 'type': 'starts-with-alnum', 'value': 'C'},
                                                                        {'name': 'label_alt_id', 'type': 'enum',
                                                                         'enum': (repAltId,)},
                                                                        {'name': 'type_symbol', 'type': 'str', 'value': 'C'}])

                            bb_atom_sites = self.getDictListWithFilter('atom_site',
                                                                       [{'name': 'Cartn_x', 'type': 'float', 'alt_name': 'x'},
                                                                        {'name': 'Cartn_y', 'type': 'float', 'alt_name': 'y'},
                                                                        {'name': 'Cartn_z', 'type': 'float', 'alt_name': 'z'},
                                                                        {'name': 'label_asym_id', 'type': 'str', 'alt_name': 'chain_id'},
                                                                        {'name': 'auth_seq_id', 'type': 'int', 'alt_name': 'seq_id'},
                                                                        {'name': 'ndb_model' if alias else 'pdbx_PDB_model_num', 'type': 'int', 'alt_name': 'model_id'},
                                                                        {'name': 'type_symbol', 'type': 'str', 'alt_name': 'element'}
                                                                        ],
                                                                       [{'name': 'label_asym_id', 'type': 'enum',
                                                                         'enum': polyPeptideChains},
                                                                        {'name': 'label_atom_id', 'type': 'starts-with-alnum', 'value': 'N'},
                                                                        {'name': 'label_alt_id', 'type': 'enum',
                                                                         'enum': (repAltId,)},
                                                                        {'name': 'type_symbol', 'type': 'str', 'value': 'N'}])

                            bb_atom_sites.extend(ca_atom_sites)
                            bb_atom_sites.extend(co_atom_sites)

                            caRmsd, caWellDefinedRegion = self.__calculateRmsd(polyPeptideChains, polyPeptideLengths,
                                                                               totalModels, effModelIds,
                                                                               ca_atom_sites, bb_atom_sites, randomM)

                        if caRmsd is not None:
                            ent['ca_rmsd'] = caRmsd[polyPeptideChains.index(c)]
                        if caWellDefinedRegion is not None:
                            ent['well_defined_region'] = caWellDefinedRegion[polyPeptideChains.index(c)]

                    elif 'ribonucleotide' in etype:

                        p_atom_sites = self.getDictListWithFilter('atom_site',
                                                                  [{'name': 'Cartn_x', 'type': 'float', 'alt_name': 'x'},
                                                                   {'name': 'Cartn_y', 'type': 'float', 'alt_name': 'y'},
                                                                   {'name': 'Cartn_z', 'type': 'float', 'alt_name': 'z'},
                                                                   {'name': 'label_asym_id', 'type': 'str', 'alt_name': 'chain_id'},
                                                                   {'name': 'auth_seq_id', 'type': 'int', 'alt_name': 'seq_id'},
                                                                   {'name': 'ndb_model' if alias else 'pdbx_PDB_model_num', 'type': 'int', 'alt_name': 'model_id'},
                                                                   {'name': 'type_symbol', 'type': 'str', 'alt_name': 'element'}
                                                                   ],
                                                                  [{'name': 'label_asym_id', 'type': 'str', 'value': c},
                                                                   {'name': 'label_atom_id', 'type': 'starts-with-alnum', 'value': 'P'},
                                                                   {'name': 'label_alt_id', 'type': 'enum',
                                                                    'enum': (repAltId,)},
                                                                   {'name': 'type_symbol', 'type': 'str', 'value': 'P'}])

                        bb_atom_sites = self.getDictListWithFilter('atom_site',
                                                                   [{'name': 'Cartn_x', 'type': 'float', 'alt_name': 'x'},
                                                                    {'name': 'Cartn_y', 'type': 'float', 'alt_name': 'y'},
                                                                    {'name': 'Cartn_z', 'type': 'float', 'alt_name': 'z'},
                                                                    {'name': 'label_asym_id', 'type': 'str', 'alt_name': 'chain_id'},
                                                                    {'name': 'auth_seq_id', 'type': 'int', 'alt_name': 'seq_id'},
                                                                    {'name': 'ndb_model' if alias else 'pdbx_PDB_model_num', 'type': 'int', 'alt_name': 'model_id'},
                                                                    {'name': 'type_symbol', 'type': 'str', 'alt_name': 'element'}
                                                                    ],
                                                                   [{'name': 'label_asym_id', 'type': 'str', 'value': c},
                                                                    {'name': 'label_atom_id', 'type': 'enum',
                                                                     'enum': ("C5'", "C4'", "C3'")},
                                                                    {'name': 'label_alt_id', 'type': 'enum',
                                                                     'enum': (repAltId,)},
                                                                    {'name': 'type_symbol', 'type': 'str', 'value': 'C'}])

                        bb_atom_sites.extend(p_atom_sites)

                        pRmsd, pWellDefinedRegion = self.__calculateRmsd([c], [len(_seqDict[c])],
                                                                         totalModels, effModelIds,
                                                                         p_atom_sites, bb_atom_sites, randomM)

                        if pRmsd is not None:
                            ent['p_rmsd'] = pRmsd[0]
                        if pWellDefinedRegion is not None:
                            ent['well_defined_region'] = pWellDefinedRegion[0]

            if len(chainIds) > 1:
                identity = []
                for _c in chainIds:
                    if _c == c:
                        continue
                    if compDict[_c] == compDict[c]:
                        identity.append(_c)
                if len(identity) > 0:
                    ent['identical_chain_id'] = identity
                    if auth_chain_id_col != -1:
                        ent['identical_auth_chain_id'] = [authChainDict[c] for c in identity]

                if len(unmapSeqIds) > 0 and c in unmapSeqIds and c in chainIdWoDefault:
                    ent['unmapped_seq_id'] = [int(s) for s, r in unmapSeqIds[c]]
                    ent['unmapped_auth_seq_id'] = [int(s) for s in unmapAuthSeqIds[c]]

            asm.append(ent)

        return asm

    def __extractStructConf(self, chain_id: str, seq_ids: List[int], label_scheme: bool = True) -> List[Optional[str]]:
        """ Extract structure conformational annotations.
        """

        ret = [None] * len(seq_ids)

        helix_id_name = 'pdbx_PDB_helix_id' if self.hasItem('struct_conf', 'pdbx_PDB_helix_id') else 'pdb_id'

        struct_conf = self.getDictListWithFilter('struct_conf',
                                                 [{'name': 'conf_type_id', 'type': 'str'},
                                                  {'name': helix_id_name, 'type': 'str', 'alt_name': 'helix_id'},
                                                  {'name': 'beg_label_seq_id' if label_scheme else 'beg_auth_seq_id', 'type': 'int', 'alt_name': 'beg_seq_id'},
                                                  {'name': 'end_label_seq_id' if label_scheme else 'end_auth_seq_id', 'type': 'int', 'alt_name': 'end_seq_id'}
                                                  ],
                                                 [{'name': 'beg_label_asym_id', 'type': 'str', 'value': chain_id},
                                                  {'name': 'end_label_asym_id', 'type': 'str', 'value': chain_id}
                                                  ])

        for sc in struct_conf:
            for seq_id in range(sc['beg_seq_id'], sc['end_seq_id'] + 1):
                if seq_id in seq_ids and sc['conf_type_id'] is not None and sc['helix_id'] is not None:
                    ret[seq_ids.index(seq_id)] = sc['conf_type_id'] + ':' + sc['helix_id']

        struct_sheet_range = self.getDictListWithFilter('struct_sheet_range',
                                                        [{'name': 'sheet_id', 'type': 'str'},
                                                         {'name': 'id', 'type': 'str'},
                                                         {'name': 'beg_label_seq_id' if label_scheme else 'beg_auth_seq_id', 'type': 'int', 'alt_name': 'beg_seq_id'},
                                                         {'name': 'end_label_seq_id' if label_scheme else 'end_auth_seq_id', 'type': 'int', 'alt_name': 'end_seq_id'}
                                                         ],
                                                        [{'name': 'beg_label_asym_id', 'type': 'str', 'value': chain_id},
                                                         {'name': 'end_label_asym_id', 'type': 'str', 'value': chain_id}
                                                         ])

        for ssr in struct_sheet_range:
            for seq_id in range(ssr['beg_seq_id'], ssr['end_seq_id'] + 1):
                if seq_id in seq_ids and ssr['sheet_id'] is not None and ssr['id'] is not None:
                    ret[seq_ids.index(seq_id)] = 'STRN:' + ssr['sheet_id'] + ':' + ssr['id']

        return ret

    def __calculateRmsd(self, chain_ids: List[str], lengths: List[int], total_models: int = 1, eff_model_ids: Optional[List[str]] = None,
                        atom_sites: Optional[List[dict]] = None, bb_atom_sites: Optional[List[dict]] = None,
                        randomM: Optional[List[list]] = None) -> Tuple[Optional[List[dict]], Optional[List[dict]]]:
        """ Calculate RMSD of alpha carbons/phosphates in the ensemble.
        """

        if None in (atom_sites, bb_atom_sites):
            return None, None

        _atom_site_dict = {}
        for model_id in eff_model_ids:
            _atom_site_dict[model_id] = [a for a in atom_sites if a['model_id'] == model_id]

        _bb_atom_site_dict = {}
        for model_id in eff_model_ids:
            _bb_atom_site_dict[model_id] = [a for a in bb_atom_sites if a['model_id'] == model_id]

        size = len(_atom_site_dict[1])

        if size == 0:
            return None, None

        matrix_size = (size, size)

        d_avr = np.zeros(matrix_size, dtype=float)

        _total_models = 0

        for model_id in eff_model_ids:

            _atom_site = _atom_site_dict[model_id]

            if len(_atom_site) == 0:
                continue

            _total_models += 1

            for a_i, a_j in itertools.combinations(_atom_site, 2):

                i = _atom_site.index(a_i)
                j = _atom_site.index(a_j)

                d = np.linalg.norm(to_np_array(a_i) - to_np_array(a_j))

                if i < j:
                    d_avr[i, j] += d
                else:
                    d_avr[j, i] += d

        if _total_models <= 1:
            return None, None

        factor = 1.0 / _total_models

        d_avr = np.multiply(d_avr, factor)

        d_var = np.zeros(matrix_size, dtype=float)

        for model_id in eff_model_ids:

            _atom_site = _atom_site_dict[model_id]

            if len(_atom_site) == 0:
                continue

            for a_i, a_j in itertools.combinations(_atom_site, 2):

                i = _atom_site.index(a_i)
                j = _atom_site.index(a_j)

                d = np.linalg.norm(to_np_array(a_i) - to_np_array(a_j))

                if i < j:
                    d -= d_avr[i, j]
                    d_var[i, j] += d * d
                else:
                    d -= d_avr[j, i]
                    d_var[j, i] += d * d

        d_var = np.multiply(d_var, factor)

        max_d_var = min(np.max(d_var), self.__d_cutoff)

        d_ord = np.ones(matrix_size, dtype=float)

        if max_d_var > 0.0:

            for i, j in itertools.combinations(range(size), 2):

                if i < j:
                    q = max(1.0 - math.sqrt(d_var[i, j] / max_d_var), 0.0)
                else:
                    q = max(1.0 - math.sqrt(d_var[j, i] / max_d_var), 0.0)

                d_ord[i, j] = d_ord[j, i] = q

        _, v = np.linalg.eig(d_ord)

        md5_set = set()

        abort = False

        min_score = 1000000.0
        min_result = None

        stop_min_samples = -1

        for min_samples in reversed(range(self.__min_samples_for_clustering, self.__max_samples_for_clustering + 1)):

            if min_samples == stop_min_samples:
                break

            for features in range(self.__min_features_for_clustering, self.__max_features_for_clustering + 1):

                x = np.delete(v, np.s_[features:], 1)

                if min_samples >= features:
                    continue

                for _epsilon in range(4, 11):

                    if abort:
                        break

                    epsilon = 2.0 ** (_epsilon / 2.0) / 100.0  # epsilon travels from 0.04 to 0.32

                    try:
                        db = DBSCAN(eps=epsilon, min_samples=min_samples).fit(x)
                    except ValueError:
                        db = DBSCAN(eps=epsilon, min_samples=min_samples).fit(np.real(x))

                    labels = db.labels_

                    list_labels = list(labels)
                    set_labels = set(labels)

                    n_clusters = len(set_labels) - (1 if -1 in set_labels else 0)
                    n_noise = list_labels.count(-1)

                    if n_clusters == 0 or n_clusters >= features - 2:
                        continue

                    md5 = hashlib.md5(str(list_labels).encode('utf-8'))

                    if md5 in md5_set:
                        continue

                    md5_set.add(md5)

                    result = {'features': features, 'min_samples': min_samples, 'epsilon': epsilon, 'clusters': n_clusters, 'noise': n_noise}

                    score = 0.0

                    for label in set_labels:

                        monomers = list_labels.count(label)

                        if monomers < self.__min_monomers_for_domain:
                            continue

                        fraction = float(monomers) / size

                        _rmsd = []

                        _atom_site_ref = _atom_site_dict[1]
                        _atom_site_p = [_a for _a, _l in zip(_atom_site_ref, list_labels) if _l == label]

                        # """
                        # if label != -1:
                        #     for chain_id in chain_ids:
                        #         seq_ids = sorted(set(a['seq_id'] for a in _atom_site_p if a['chain_id'] == chain_id))
                        #         if len(seq_ids) > 0:
                        #             gaps = seq_ids[-1] + 1 - seq_ids[0] - len(seq_ids)

                        #             if gaps > monomers:
                        #                 score = 0.0
                        #                 break
                        # """

                        for model_id in range(2, total_models + 1):

                            if model_id not in eff_model_ids:
                                continue

                            _atom_site_test = _atom_site_dict[model_id]

                            if len(_atom_site_test) == 0:
                                continue

                            _atom_site_q = [_a for _a, _l in zip(_atom_site_test, list_labels) if _l == label]

                            _rmsd.append(calculate_rmsd(_atom_site_p, _atom_site_q))

                        mean_rmsd = np.mean(np.array(_rmsd))

                        score += mean_rmsd * fraction

                    if score == 0.0:
                        continue

                    result['score'] = score

                    if n_clusters > 0 and stop_min_samples == -1:
                        stop_min_samples = min_samples - 2

                    if self.__verbose and self.__debug:
                        self.__lfh.write(f'{result}\n')

                    if score < min_score or (n_noise == 0 and min_score < self.__rmsd_overlaid_exactly):
                        min_score = score
                        min_result = result

                        if n_noise == 0 and min_score < self.__rmsd_overlaid_exactly:
                            abort = True

        if min_result is None:
            return None, None

        x = np.delete(v, np.s_[min_result['features']:], 1)

        try:
            db = DBSCAN(eps=min_result['epsilon'], min_samples=min_result['min_samples']).fit(x)
        except ValueError:
            db = DBSCAN(eps=min_result['epsilon'], min_samples=min_result['min_samples']).fit(np.real(x))

        labels = db.labels_

        list_labels = list(labels)
        domains = collections.Counter(list_labels).most_common()

        if domains[0][0] == -1:
            return None, None

        eff_labels = [label for label, count in domains if label != -1 and count >= self.__min_monomers_for_domain]
        eff_domain_id = {}

        n_clusters = len(set(labels)) - (1 if -1 in labels else 0)
        if self.__verbose and self.__debug:
            self.__lfh.write(f"feature: {min_result['features']}, "
                             f"min_sample: {min_result['min_samples']}, epsilon: {min_result['epsilon']}, "
                             f"clusters: {n_clusters} (effective domains: {len(eff_labels)}), score: {min_score}\n")

        _chain_ids = [a['chain_id'] for a in _atom_site_dict[1]]
        _seq_ids = [a['seq_id'] for a in _atom_site_dict[1]]

        domain_id = 1
        for label, chain_id, seq_id in zip(labels, _chain_ids, _seq_ids):
            if label not in eff_labels:
                if self.__verbose and self.__debug:
                    self.__lfh.write(f"chain_id: {chain_id}, seq_id: {seq_id}, domain_id: -1\n")
            else:
                _label = int(label)
                if _label not in eff_domain_id:
                    eff_domain_id[_label] = domain_id
                    domain_id += 1
                if self.__verbose and self.__debug:
                    self.__lfh.write(f"chain_id: {chain_id}, seq_id: {seq_id}, domain_id: {eff_domain_id[_label]} label: {_label}\n")

        rlist = []
        for chain_id in chain_ids:
            rlist.append([])

        for ref_model_id in eff_model_ids:

            item = {'model_id': ref_model_id}
            _atom_site_ref = _atom_site_dict[ref_model_id]
            _bb_atom_site_ref = _bb_atom_site_dict[ref_model_id]

            dst_chain_ids = None
            min_label = -1
            min_core_rmsd = mean_align_rmsd = 1000000.0

            for label, count in domains:

                if label not in eff_labels:
                    break

                _label = int(label)

                _atom_site_p = [_a for _a, _l in zip(_atom_site_ref, list_labels) if _l == label]

                _dst_chain_ids = set(_a['chain_id'] for _a in _atom_site_p)
                _seq_keys = sorted(set((_a['chain_id'], _a['seq_id']) for _a in _atom_site_p),
                                   key=itemgetter(0, 1))
                _bb_atom_site_p = [_a for _a in _bb_atom_site_ref if (_a['chain_id'], _a['seq_id']) in _seq_keys]

                core_rmsd, align_rmsd, exact_overlaid_model_ids = [], [], []

                for test_model_id in eff_model_ids:

                    if ref_model_id == test_model_id:
                        continue

                    _atom_site_test = _atom_site_dict[test_model_id]
                    _atom_site_q = [_a for _a, _l in zip(_atom_site_test, list_labels) if _l == label]

                    _bb_atom_site_test = _bb_atom_site_dict[test_model_id]
                    _bb_atom_site_q = [_a for _a in _bb_atom_site_test if (_a['chain_id'], _a['seq_id']) in _seq_keys]

                    _core_rmsd = []

                    for idx in range(count):

                        ref_atom = _atom_site_p[idx]
                        test_atom = _atom_site_q[idx]

                        ref_v = to_np_array(ref_atom)
                        if self.__random_rotaion_test:
                            ref_v = np.dot(randomM[ref_model_id], ref_v)

                        test_v = to_np_array(test_atom)
                        if self.__random_rotaion_test:
                            test_v = np.dot(randomM[test_model_id], test_v)
                        d = test_v - ref_v
                        _core_rmsd.append(np.dot(d, d))

                    core_rmsd.append(math.sqrt(np.mean(np.array(_core_rmsd))))
                    _rmsd = calculate_rmsd(_bb_atom_site_p, _bb_atom_site_q)
                    align_rmsd.append(_rmsd)
                    if _rmsd < self.__rmsd_overlaid_exactly and ref_model_id < test_model_id:
                        exact_overlaid_model_ids.append({'ref_model_id': ref_model_id,
                                                         'test_model_id': test_model_id,
                                                         'rmsd_in_well_defined_region': float(f"{_rmsd:.4f}")})

                mean_core_rmsd = np.mean(np.array(core_rmsd))

                if mean_core_rmsd < min_core_rmsd:
                    dst_chain_ids = _dst_chain_ids
                    min_label = _label
                    min_core_rmsd = mean_core_rmsd
                    mean_align_rmsd = np.mean(np.array(align_rmsd))

            if min_label != -1:
                item['domain_id'] = eff_domain_id[min_label]
                item['raw_rmsd_in_well_defined_region'] = float(f"{min_core_rmsd:.4f}")
                item['rmsd_in_well_defined_region'] = float(f"{mean_align_rmsd:.4f}")
                if len(exact_overlaid_model_ids) > 0:
                    item['exactly_overlaid_model'] = exact_overlaid_model_ids
                for chain_id in dst_chain_ids:
                    rlist[chain_ids.index(chain_id)].append(item)

        dlist = []
        for chain_id in chain_ids:
            dlist.append([])

        seq_range_p = re.compile(r'^\[(-?\d+)-(-?\d+)\]$')

        for label in sorted(eff_labels):

            item = {}

            r = np.full((_total_models, _total_models), self.__rmsd_cutoff, dtype=float)

            _rmsd = []

            for ref_model_id in range(1, _total_models):

                if ref_model_id not in eff_model_ids:
                    continue

                _atom_site_ref = _atom_site_dict[ref_model_id]
                _atom_site_p = [_a for _a, _l in zip(_atom_site_ref, list_labels) if _l == label]

                _bb_atom_site_ref = _bb_atom_site_dict[ref_model_id]
                _seq_keys = sorted(set((_a['chain_id'], _a['seq_id']) for _a in _atom_site_p),
                                   key=itemgetter(0, 1))
                _bb_atom_site_p = [_a for _a in _bb_atom_site_ref if (_a['chain_id'], _a['seq_id']) in _seq_keys]

                for test_model_id in range(2, _total_models + 1):

                    if ref_model_id >= test_model_id or test_model_id not in eff_model_ids:
                        continue

                    # _atom_site_test = _atom_site_dict[test_model_id]
                    # _atom_site_q = [_a for _a, _l in zip(_atom_site_test, list_labels) if _l == label]

                    _bb_atom_site_test = _bb_atom_site_dict[test_model_id]
                    _bb_atom_site_q = [_a for _a in _bb_atom_site_test if (_a['chain_id'], _a['seq_id']) in _seq_keys]

                    _rmsd_ = calculate_rmsd(_bb_atom_site_p, _bb_atom_site_q)

                    r[ref_model_id - 1, test_model_id - 1] = _rmsd_
                    r[test_model_id - 1, ref_model_id - 1] = _rmsd_

                    _rmsd.append(_rmsd_)

            if len(_rmsd) > 0:
                item['mean_rmsd'] = float(f"{np.mean(np.array(_rmsd)):.4f}")

            _, v = np.linalg.eig(r)
            x = np.delete(np.abs(v), np.s_[1:], 1)
            ref_model_id = int(np.argmin(x, axis=0)[0]) + 1

            item['medoid_model_id'] = ref_model_id

            _atom_site_ref = _atom_site_dict[ref_model_id]
            _atom_site_p = [_a for _a, _l in zip(_atom_site_ref, list_labels) if _l == label]
            _bb_atom_site_ref = _bb_atom_site_dict[ref_model_id]
            _seq_keys = sorted(set((_a['chain_id'], _a['seq_id']) for _a in _atom_site_p))
            _bb_atom_site_p = [_a for _a in _bb_atom_site_ref if (_a['chain_id'], _a['seq_id']) in _seq_keys]

            _rmsd = []

            for test_model_id in eff_model_ids:

                if ref_model_id == test_model_id:
                    continue

                # _atom_site_test = _atom_site_dict[test_model_id]
                # _atom_site_q = [_a for _a, _l in zip(_atom_site_test, list_labels) if _l == label]

                _bb_atom_site_test = _bb_atom_site_dict[test_model_id]
                _bb_atom_site_q = [_a for _a in _bb_atom_site_test if (_a['chain_id'], _a['seq_id']) in _seq_keys]

                _rmsd.append(calculate_rmsd(_bb_atom_site_p, _bb_atom_site_q))

            if len(_rmsd) > 0:
                item['medoid_rmsd'] = float(f"{np.mean(np.array(_rmsd)):.4f}")

            _item = copy.copy(item)

            _label = int(label)

            dst_chain_ids = set(a['chain_id'] for a, l in zip(_atom_site_ref, list_labels) if l == label)  # noqa: E741
            seq_keys = sorted(set((a['chain_id'], a['seq_id']) for a, l in zip(_atom_site_ref, list_labels) if l == label),  # noqa: E741
                              key=itemgetter(0, 1))

            for chain_id in dst_chain_ids:
                seq_ids = [s for c, s in seq_keys if c == chain_id]
                count = len(seq_ids)

                _item['domain_id'] = eff_domain_id[_label]
                _item['number_of_monomers'] = count
                _item['seq_id'] = seq_ids
                gaps = seq_ids[-1] + 1 - seq_ids[0] - len(seq_ids)
                _item['number_of_gaps'] = gaps
                if gaps == 0:
                    seq_range = f"[{seq_ids[0]}-{seq_ids[-1]}]"
                else:
                    seq_range = f"[{seq_ids[0]}-"
                    for idx, seq_id in enumerate(seq_ids):
                        if idx > 0 and seq_id != seq_ids[idx - 1] + 1:
                            seq_range += f"{seq_ids[idx - 1]}],[{seq_id}-"
                    seq_range += f"{seq_ids[-1]}]"
                _seq_range = []
                for r in seq_range.split(','):
                    try:
                        g = seq_range_p.search(r).groups()
                        if g[0] != g[1]:
                            _seq_range.append(f"[{g[0]}-{g[1]}]")
                        else:
                            _seq_range.append(f"[{g[0]}]")
                    except AttributeError:
                        pass
                _item['range_of_seq_id'] = ','.join(_seq_range)
                _item['percent_of_core'] = float(f"{float(count) / lengths[chain_ids.index(chain_id)] * 100.0:.1f}")

                dlist[chain_ids.index(chain_id)].append(_item)

        if self.__verbose and self.__debug:
            self.__lfh.write(f"{dlist}\n")

        return rlist, dlist
